        page    60, 132
        title   Treiber fr 2 COMs mit Int. (Vers. 3.31) In/Out

;-------------------------------------------------------------------------
;  J. Rathlev, Jun. 1995
;  Vers. 3.2 mit Interrupt-Pufferung der Ein- und Ausgabe (95)
;  Vers. 3.3 mit Zeilenzhlung bei <CR>
;  Vers. 3.4 mit Zeilenfilter (Mai 1999)
;-------------------------------------------------------------------------
;  Das nachfolgende Programm emuliert den BIOS-Interrupt 14
;  (comm. port) unter  Benutzung eines interruptgesteuerten
;  Buffers fr den Datenempfang fr zwei Kanle (z.B. COM2-3).
;  Durch ndern der Konstanten COMANZ auch fr andere Anzahl von COMs
;  verwendbar (Achtung: ev. Interruptroutine erweitern)
;  Portadressen: beliebig
;  Interrupts:   1 .. 15
;  Die alten Funktionen
;               ah = 0 : Initialisierung (Baud Rate etc. und siehe Fkt. 4)
;                  = 1 : Daten senden
;                  = 2 : Daten empfangen
;                  = 3 : Status lesen
;  sind erhalten geblieben.
;  Hinzu ist gekommen
;               ah = 4 : Modem einschalten, Buffer Init. und Interrupt an
;                        (auch in Funktion 0 enthalten)
;                  = 5 : Modem ausschalten, Interrupt aus
;                  = 6 : Line Control Register auslesen
;                  = 7 : Empf.-/Sendebuffer Counter auslesen
;                  = 8 : Treiber Identifizierung
;
;  Zeilenfilter (Rev 3.4): Es kann beim Laden des Programms eingestellt
;               werden, ob an der Schnittstelle ankommende Zeilen
;               berlesen werden sollen. Standardmig wird jede
;               Zeile bernommen, bei anderer Einstellung nur jede
;               n-te (z.B. nur jede zehnte). Diese Mglichkeit wurde
;               eingebaut, um zu verhindern, da Anwenderprogramme
;               durch Datenquellen, die zu viele Daten senden, zugestopft
;               werden.
;
;  Das Programm wird nach dem Laden des Betriebssystem MS-DOS einmal
;  aufgerufen (z.B. mit "autoexec.bat") und bleibt dann im Speicher
;  resident:
;  Command-String:
;     COM2DRV      ==>  COM2-Treiber installieren (Standard)
;                       COM1: Port 3F8, IRQ 4, jede Zeile
;                       COM2: Port 2F8, IRQ 3, jede Zeile
;     COM2DRV [n] /Pport1,port2 /Iirq1,irq2 /Lln1,ln2
;                  ==>  COM2-Treiber installieren
;                       mit alternativen Portadressen (portx),
;                       Interrupts (irqx = 1 .. 15)
;                       und Zeilenfilter (lnx)
;     COM2DRV n    ==>  COM2-Treiber auf COMn und COM(n+1) inst.
;     COM2DRV /u   ==>  COM2-Treiber wieder entfernen
;
;     Bsp.:  COM2DRV 2 /P2F8,3E8 /I3,9
;
;  Alle Aufrufe von Int 14 durch irgendein Programm werden
;  abgefangen und durch die unten beschriebenen Routinen bedient.
;
;  Hardware-Voraussetzungen:
;       1. Interface fr COM1 und COM2 (bzw. COMn und COM(n+1))
;       2. Die Jumper auf der Platine mssen auf die bei der Option /I
;          angegebenen Interrupts gesetzt sein (Standard IRQ4 und 3).
;       3. Die Schalterstellung fr die Portadressen mu der Angabe
;          nach der Option /P enstprechen (Standard 3F8 und 2F8)
;
;  Schnittstellen-Protokoll:
;    Senden: Die Funktion 0 oder 4 (Modem an) gibt die
;       Signale DTR (pin 20) und RTS (pin 4) aus (+ 12 V).
;       Vor dem Senden eines Zeichens werden DSR (pin 6) und CTS (pin 5)
;       abgefragt. Sie mssen beide gesetzt sein (+12 V). Wenn diese
;       Bedingung nicht erfllt ist, erfolgt ein Rcksprung aus der
;       Senderoutine (Funktion 1) mit dem Time-out-bit (ah-bit 7)
;       auf 1 gesetzt.
;    Empfangen: Es ist grundstzlich DTR (pin 20) gesetzt. Alle ankommenden
;       Zeichen werden im Buffer zwischengespeichert (sofern mit Funktion 0
;       oder 4 das Modem eingeschaltet wurde).
;       Der Buffer kann max. 256 Zeichen pro Schnittstelle mit dem
;       zugehrigen Fehlerstatus (Overrun-, Parity- oder Framing-Error)
;       aufnehmen. Weitere Zeichen werden ignoriert.
;
;  Status-Abfrage:
;       Die Funktion 3 liest den gleichen Status wie im Original-
;       BIOS mit Ausnahme des DR-Bits (Data Ready, ah-bit 0). Dieses Bit
;       zeigt hier an, da im Input-Buffer mindestens ein Zeichen gespeichert
;       ist, das mit der Funktion 2 abgeholt werden kann.
;
;-------------------------------------------------------------------------


code    segment
        assume  cs:code
        assume  ds:code
        assume  es:nothing
        assume  ss:nothing


; BIOS Interrupt Number

BioInt  equ     14h

; Number of COM-channels

ComAnz  equ     2

CR      equ     0DH     ;carriage return

; Registers of 8250:

IER     equ     1       ; Interrupt Enable Register
IIR     equ     2       ; Interrupt Identification Register
LCR     equ     3       ; Line Control Register
MCR     equ     4       ; MODEM Control Register
LSR     equ     5       ; Line Status Register
MSR     equ     6       ; MODEM Status Register

EOI     equ     20h     ; end of interrupt

;-------------------------------------------------------------------------
        org     100h

start:  jmp     init

DrvName : db 'COM2DRV.COM (3.4)',0
CopRgt  : db 'PC-COM(2)-Driver (Vers. 3.40 - (c) J.Rathlev 1999) $'

;Time out counter for data output

OneSec  equ 18          ;timer clock is 18 per sec

ToCnt:  dw      0


; input buffer

RcvLen    equ     256           ;receiver buffer length
SndLen    equ     512           ;send buffer length

RcvBuf  dw      ComAnz*RcvLen dup (?)
SndBuf  db      ComAnz*SndLen dup (?)

; bufferpointer and counter
; receive buffer
RcvBPtr  label   word
        n = 0
rept    ComAnz
        dw      RcvBuf+n
        n = n + 2*RcvLen
        endm

RcvCnt    dw      ComAnz dup (?)       ;buffer count
LineCnt   dw      ComAnz dup (?)       ;line count
LiFiCnt   dw      ComAnz dup (?)       ;line filter
RcvIPnt   dw      ComAnz dup (?)       ;input buffer pointer
RcvOPnt   dw      ComAnz dup (?)       ;output buffer pointer

; send buffer
SndBPtr  label   word
        n = 0
rept    ComAnz
        dw      SndBuf+n
        n = n + SndLen
        endm

SndCnt    dw      ComAnz dup (?)       ;buffer count
SndIPnt   dw      ComAnz dup (?)       ;input buffer pointer
SndOPnt   dw      ComAnz dup (?)       ;output buffer pointer

TmPtr     dd      ?                    ;pointer to old timer interrupt
HwiPtr    dd      ComAnz dup (?)       ;Pointer to old ser. Hardware Int.

PortAdr   dw      ComAnz dup (?)       ;RS232 port Adr.
HwInt     dw      ComAnz dup (?)       ;IRQ Number and Mask
LiFiSet   dw      ComAnz dup (0)       ;line filter setting

LastCom   dw      ComAnz-1             ;last COM
FrstCom   dw      0                    ;first COM
IntPtr    dd      ?                    ;Pointer to BIOS-INT14
ComDrvID  dw      0F234h               ;driver identifier byte

;-----------------------------------------------------------------
; the following routine "ComRout" is the new INT14 for COMx
; port number in dx
; function in ah:
;-------------------------------------------------------------------------
;  0    initialize the comm. port
;       enable interrupts from line 4
;       al contains parameter for init.:
;       normal: (bit 1 = 1)
;       bit
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -    110         X0 - none    0 - 1   10 - 7 bits
;       001 -    150         01 - odd     1 - 2   11 - 8 bits
;       010 -    300         11 - even
;       011 -    600
;       100 -   1200
;       101 -   2400
;       110 -   4800
;       111 -   9600
;       on return status in ah,al as function 3
;
;       extended: (bit 1 = 0)
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -     50         X0 - none    0 - 1   00 - 7 bits
;       001 -     75         01 - odd     1 - 2   01 - 8 bits
;       010 -    134.5       11 - even
;       011 -   1800
;       100 -  19200
;       101 -  38400
;       110 -  57600
;       111 - 115200
;
;       on return status in ah,al as function 3
;
;-------------------------------------------------------------------------
;  1    send a char. in al to comm. line
;       al is saved
;       return with
;          ah bit 7 = 1 : unable to send char. to line
;                   = 0 : bit 0-6 contain line status
;
;-------------------------------------------------------------------------
;  2    read a char. to al from comm. line via buffer
;       return with
;          ah bit 7 = 1 : no char. in buffer
;                   = 0 : bit 1,2,3,4 contain line status
;
;-------------------------------------------------------------------------
;  3    read status from comm. port
;       return with
;          ah bit 7 : always 0
;                 6 : transmitter shift reg. empty
;                 5 : transmitter holding reg. empty
;                 4 : break interrupt
;                 3 : framing error
;                 2 : parity error
;                 1 : overrun error
;                 0 : data received (in buffer)
;          al bit 7 : received line signal detect
;                 6 : ring indicator
;                 5 : data set ready
;                 4 : clear to send
;                 3 : delta received line signal detect
;                 2 : trailing edge ring indicator
;                 1 : delta data set ready
;                 0 : delta clear to send
;
;-------------------------------------------------------------------------
;  4    switch modem on (set DTR and RTS, enable interrupts)
;
;-------------------------------------------------------------------------
;  5    switch modem off (reset DTR and RTS, interrupt disable)
;
;-------------------------------------------------------------------------
;  6    read line control register to al
;
;-------------------------------------------------------------------------
;  7    read buffer counter to ax, buffer length to dx
;       line count to cx
;
;-------------------------------------------------------------------------
;  8    driver identifikation: set al=0 on call
;                              dual COM-driver installed:
;                                     return al=$F2
;                              otherwise al unchanged
;-------------------------------------------------------------------------

ComRout:sti
        push    ds
        push    si
        push    di
        mov     si,cs           ;copy cs to ds
        mov     ds,si
        cmp     dx,FrstCom      ;test if COM-Nr. is supported
        jl      Direct
        cmp     dx,LastCom      ;last COM-Nr. supported
        jg      Direct          ;above, jump to BIOS INT14
        sub     dx,FrstCom
        push    bx
        push    cx
        push    dx
        mov     bl,ah           ;function to bl
        xor     bh,bh
        shl     bx,1            ;*2
        cmp     bx,FctNr        ;legal function?
        jnc     ComFin          ;no, return imm.
        shl     dx,1
        mov     si,dx           ;2*ComNr.
        mov     dx,[si+PortAdr] ;Portaddr.
        jmp     word ptr [bx+offset FctTab]   ;branch

FctTab  dw      ComSet          ;set baud rate and format
        dw      ComXmit         ;send char. via ser. port
        dw      ComRecv         ;get a received char. from buffer
        dw      ComStat         ;get modem status
        dw      ComOn           ;switch modem on
        dw      ComOff          ;switch modem off
        dw      ComLCR          ;read LCR
        dw      ComBCnt         ;read buffer counter
        dw      ComID           ;driver identifier
FctNr   equ     $-FctTab

;-------------------------------------------------------------------------
ComFin: pop     dx
        pop     cx
        pop     bx
        jmp     ComRet


;jump to BIOS-INT14 routine

Direct: pushf           ;prepare to emulate int-call
        cli
        call    dword ptr [IntPtr]    ;call to BIOS
ComRet: pop     di
        pop     si
        pop     ds
        iret


;-------------------------------------------------------------------------
;baud rate table

BTab1   dw      1047            ;110 baud
        dw      768             ;150 baud
        dw      384             ;300 baud
        dw      192             ;600 baud
        dw      96              ;1200 baud
        dw      48              ;2400 baud
        dw      24              ;4800 baud
        dw      12              ;9600 baud
BTab2   dw      2304            ;50 baud
        dw      1536            ;75 baud
        dw      857             ;134.5 baud
        dw      64              ;1800 baud
        dw      6               ;19200 baud
        dw      3               ;38400 baud
        dw      2               ;57600 baud
        dw      1               ;115200 baud

;-------------------------------------------------------------------------
;set ser. port to new baud rate and format in ax
;si = 2*port nr., dx = port addr., ds = cs
;new values in al

ComSet: push    ax              ;save
        call    ComRlse
        pop     ax
        push    dx              ;save port addr.
        mov     ah,al           ;save
        add     dx,LCR          ;line control reg.
        mov     al,80H          ;set DLAB
        out     dx,al
        mov     bl,ah
        mov     cl,4            ;rot count
        rol     bl,cl           ;rotate 4 bits
        and     bx,0EH          ;mask out baud rate
        test    ah,2            ;extended mode?
        jz      ComExt
        add     bx,offset BTab1 ;pointer to 1. divisor table
        jmp     short ComNor
ComExt: add     bx,offset BTab2 ;pointer to 2. divisor table
ComNor: dec     dx
        dec     dx              ;divisor latch (DLM)
        mov     al,[bx+1]
        out     dx,al
        dec     dx              ;divisor latch (DLL)
        mov     al,[bx]
        out     dx,al
        add     dx,LCR          ;line control reg.
        mov     al,ah           ;format bits
        or      al,2            ;set bit 1
        and     al,1FH          ;mask out
        out     dx,al
        pop     dx              ;restore port base addr.
        call    ComInit         ;initialize
        jmp     ComStat


;-------------------------------------------------------------------------
;wait until char. received and read char. from buffer
;si = 2*port nr., dx = port addr., ds = cs
;return with char. in al, ah = 80h on time out, else = Line Status

ComRecv:lea     bx,[si+RcvCnt]    ;test buffer count
        xor     cx,cx
ComRWt: cmp     word ptr [bx],0
        jnz     ComRc2          ;char. received
        loop    ComRWt          ;wait until not zero
        mov     ah,80H          ;time out
        jmp     ComFin          ;return imm.

ComRc2: cli
        dec     word ptr [bx]   ;count one down
        mov     bx,[si+RcvOPnt]   ;buffer pointer
        mov     di,[si+RcvBPtr]   ;buffer base addr.
        mov     ax,[bx+di]      ;get status and char. from buffer
        inc     bx              ;incr. buffer pointer
        inc     bx
        cmp     bx,2*RcvLen       ;buffer length
        jc      ComRc1
        xor     bx,bx           ;set to first buffer addr.
ComRc1: mov     [si+RcvOPnt],bx   ;and store
        sti
        cmp     al,CR           ;carriage return?
        jnz     ComRc3
        cmp     word ptr [si+LineCnt],0
        jz      ComRc3
        dec     word ptr [si+LineCnt]
ComRc3: jmp     ComFin


;-------------------------------------------------------------------------
;send char. in al via ser. port
;si = 2*port nr., dx = port addr., ds = cs
;return with ah (bit 7) = 1 buffer full, al = char
;            ah (bit 7) = 0 on success, al =char

ComXmit:lea     bx,[si+SndCnt]    ;address of buffer counter
        cmp     word ptr [bx],SndLen
        jge     ComXm1            ;buffer full

        cli
        or      word ptr [bx],0
        jnz     ComXm3            ;interrupt active
        mov     ah,al             ;save char.
        add     dx,IER
        call    SndEna            ;enable THRE interr.
        mov     al,ah
ComXm3: inc     word ptr [bx]     ;incr. buffer count
        mov     bx,[si+SndIPnt]   ;buffer pointer
        mov     di,[si+SndBPtr]   ;buffer base addr.
        mov     [bx+di],al        ;store to buffer
        inc     bx                ;incr. buffer pointer
        cmp     bx,SndLen         ;buffer length
        jc      ComXm2
        xor     bx,bx             ;set to first buffer addr.
ComXm2: mov     [si+SndIPnt],bx   ;and store
        sti
        xor     ah,ah
        jmp     comfin

ComXm1: mov     ah,80H          ;set error bit
        jmp     ComFin


;-------------------------------------------------------------------------
;read status from ser. port
;si = 2*port nr., dx = port addr., ds = cs
;ah and al contain the status bits like the orig. BIOS-Int.
;except bit 0 in ah (data ready). This bit is set if the input buffer
;is not empty (RcvCnt)

ComStat:add     dx,LSR          ;line status reg.
        in      al,dx
        mov     ah,al
        inc     dx              ;modem status reg.
        in      al,dx
        and     ah,0FEH         ;mask out bit 0
        mov     bx,[si+RcvCnt]
        or      bx,bx
        jz      ComSt1
        or      ah,1            ;set bit 0
ComSt1: jmp     ComFin


;-------------------------------------------------------------------------
;switch modem on
;si = 2*port nr., dx = port addr., ds = cs

ComOn:  call    ComInit
        jmp     ComFin


;-------------------------------------------------------------------------
;switch modem off
;si = 2*port nr., dx = port addr., ds = cs

ComOff: call    ComRlse
        jmp     ComFin


;-------------------------------------------------------------------------
; read line control register
;si = 2*port nr., dx = port addr., ds = cs

ComLCR: add     dx,LCR
        in      al,dx
        jmp     ComFin


;-------------------------------------------------------------------------
; read buffer counter
; si = 2*port nr., dx = port addr., ds = cs
; al = 0 (Rcv)
;      1 (Snd)
; return: ax = buffer counter
;         cx = line counter (on receive mode)
;         dx = buffer length

ComBCnt:pop     dx
        pop     cx
        pop     bx
        or      al,al
        jz      GetRcv
        mov     ax,[si+SndCnt]
GetRtn: mov     dx,SndLen
        jmp     ComRet

GetRcv: mov     ax,[si+RcvCnt]
        mov     cx,[si+LineCnt]
        mov     dx,RcvLen
        jmp     ComRet


;-------------------------------------------------------------------------
; get driver identification
; al = $F2 for double COM-driver
; al = version
; cl = IRQ
; dx = port
; es:bx = pointer to driver name

ComID:  mov     ax,dx           ;save port addr.
        pop     dx
        pop     cx
        pop     bx
        lea     bx,[DrvName]
        mov     dx,ax
        mov     ax,cs
        mov     es,ax
        mov     cx,[si+HwInt]   ;interrupt
        mov     ax,ComDrvID     ;identification byte
        jmp     ComRet

;-------------------------------------------------------------------------
;subroutine test ser. port status
;in dx port addr.of status reg., in bh status bits to test
;destroyed: cx
;return with zero flag = 0 if time out, status in ah

TstStat:xor     cx,cx           ;clear
TstSta1:in      al,dx
        mov     ah,al
        xor     al,0FFH         ;complement
        test    al,bl           ;test bits
        jz      TstStaR         ;o.k. return
        loop    TstSta1         ;try again
        or      bh,bh           ;clear zero flag
TstStaR:ret


;-------------------------------------------------------------------------
;subroutine init. buffer and set modem on
;si = 2*port nr., dx = base port addr.; ds = cs

ComInit:push    dx              ;save port-addr.
        xor     ax,ax
        cli
        mov     [si+RcvCnt],ax    ;receiver buffer count
        mov     [si+LineCnt],ax   ;         line counter
        mov     [si+LiFiCnt],ax   ;         line filter
        mov     [si+RcvIPnt],ax   ;         buffer pointer
        mov     [si+RcvOPnt],ax
        mov     [si+SndCnt],ax    ;send buffer count
        mov     [si+SndIPnt],ax   ;     buffer pointer
        mov     [si+SndOPnt],ax
        in      al,dx           ;clear receiver buffer reg.
        add     dx,MCR          ;modem control reg. - modem on
        mov     al,0BH          ;set DTR, RTS and OUT2
        out     dx,al
        sub     dx,MCR-IER      ;enable interrupt
        mov     al,1            ;for receive
        out     dx,al
        mov     ax,[si+HwInt]   ;set interrupt controller mask
        not     ah
        mov     dx,21h          ;1. int. contr.
        test    al,8            ;2. ?
        jz      SetMsk
        mov     dx,0A1h         ;2. int. contr.
SetMsk: in      al,dx           ;interrupt mask
        and     al,ah           ;enable interrupt
        out     dx,al
        sti
        pop     dx
        ret


;-------------------------------------------------------------------------
;subroutine release and modem off
;dx = base port addr.; ds = cs

ComRlse:
        push    dx              ;save port-addr.
        xor     al,al
        out     dx,al           ;IER
        add     dx,3            ;MCR - set modem off
        out     dx,al
        mov     ax,[si+HwInt]   ;reset interrupt controller mask
        mov     dx,21h          ;1. int. contr.
        test    al,8            ;2. ?
        jz      ClrMsk
        mov     dx,0A1h         ;2. int. contr.
ClrMsk: in      al,dx           ;interrupt mask
        or      al,ah           ;disable interrupt
        out     dx,al
        pop     dx
        ret

;-------------------------------------------------------------------------
;enable interrupts from THRE
; dx = port addr. of IER

SndEna: in      al,dx
        or      al,2              ;enable THRE-Interr.
        out     dx,al
        ret

;-------------------------------------------------------------------------
;disable interrupt from THRE
; dx = port addr. of IER

SndDis: in      al,dx
        and     al,0Dh          ;disable THRE-Interr.
        out     dx,al
        ret

;-------------------------------------------------------------------------
;receive char. interrupt
; si =  2 * rel. ComNr
; dx =  portaddr. + IIR

assume ds:code

RcvInt: add     dx,LSR-IIR      ;Line status reg.
        in      al,dx
        and     al,1EH          ;mask out error bits
        mov     ah,al
        sub     dx,LSR
        in      al,dx           ;read char. from ser.port

        lea     bx,[si+RcvCnt]            ;address of buffer counter
        cmp     word ptr [bx],RcvLen
        jz      RcvFul                    ;buffer full

        cmp     al,CR                     ;carriage return
        mov     di,[si+LiFiCnt]
        jnz     RcvI1
        cmp     di,0
        jz      RcvI2
        dec     word ptr [si+LiFiCnt]     ;decr. filter count
        jmp     RcvFul

RcvI2:  inc     word ptr [si+LineCnt]     ;incr. line count
        mov     di,[si+LiFiSet]
        mov     [si+LiFiCnt],di           ;set filter count
        jmp     RcvI3

RcvI1:  cmp     di,0
        jnz     RcvFul                    ;ignore line

RcvI3:  inc     word ptr [bx]             ;incr. buffer count
        mov     bx,[si+RcvIPnt]           ;buffer pointer
        mov     di,[si+RcvBPtr]           ;buffer base addr.
        mov     [bx+di],ax                ;store to buffer
        inc     bx                        ;incr. pointer
        inc     bx
        cmp     bx,2*RcvLen
        jc      StBiP
        xor     bx,bx                     ;init. buffer addr.
StBiP:  mov     [si+RcvIPnt],bx           ;store

RcvFul: add     dx,IIR
        ret

;send char. interrupt

SndInt: or      word ptr [si+SndCnt],0  ;buffer counter
        jz      SndOff                  ;buffer not busy
        dec     word ptr [si+SndCnt]    ;decr. buffer count
        sub     dx,IIR
        mov     bx,[si+SndOPnt]         ;buffer pointer
        mov     di,[si+SndBPtr]         ;buffer base addr.
        mov     al,[bx+di]              ;load from buffer
        out     dx,al                   ;send char.
        inc     bx                      ;incr. pointer
        cmp     bx,SndLen
        jc      StBoP
        xor     bx,bx                   ;init. buffer addr.
StBoP:  mov     [si+SndOPnt],bx         ;store
        add     dx,IIR
        ret

SndOff: dec     dx              ;to IER

;        push    dx              ;save
;        add     dx,LSR-IER      ;to LSR
;        mov     word ptr [ToCnt],OneSec    ;set time out count
;SoLp:   in      al,dx           ;get line status
;        test    al,40h          ;test TSRE
;        jnz     SoOk            ;is empty
;        cmp     word ptr [ToCnt],0
;        jnz     SoLp            ;wait
;        pop     dx

SoOk:   call    SndDis
        inc     dx              ;to IIR
        ret


; MSR interrupt, should not happen

MSRInt: add     dx,MSR-IIR
        in      al,dx
        sub     dx,MSR-IIR
        ret

; LSR interrupt, should not happen

LSRInt: add     dx,LSR-IIR
        in      al,dx
        sub     dx,LSR-IIR
        ret

assume ds:nothing

;-------------------------------------------------------------------------
;interrupt routines for ser. port
; set si = 2 * rel. ComNr

Com2Int:
        push    ax
        push    si
        mov     si,2
        jmp     short ComInt

Com1Int:
        push    ax
        push    si
        xor     si,si

ComInt: sti
        push    bx
        push    cx
        push    dx
        push    di
        push    ds
        mov     ax,cs
        mov     ds,ax
        mov     cx,5            ;try 5 times

assume ds:code

IntLp:  mov     dx,[si+PortAdr]
        add     dx,IIR
IPoll:  in      al,dx           ;read IIR
        test    al,1            ;interrupt pending
        jnz     IntEnd          ;no
        and     ax,06h
        mov     bx,ax
        call    word ptr [bx+offset IntBra]
        jmp     short IPoll

assume ds:nothing

IntEnd: loop    IntLp           ;try again

        pop     ds
        pop     di
        pop     dx
        pop     cx
        pop     bx
        mov     ax,[si+HwInt]
        pop     si
        cli
        test    al,8
        mov     al,EOI          ;end of  interrupt
        jz      EOInt1          ;1. contr.
        out     0A0h,al         ;end of  interrupt (contr.2)
        jcxz    EOInt1          ;wait state
EOInt1: out     20h,al          ;end of  interrupt (contr.1)
        pop     ax
        iret

IntBra: dw      MSRInt
        dw      SndInt
        dw      RcvInt
        dw      LSRInt

;-------------------------------------------------------------------------
; timer interrupt

TmInt : pushf
        cmp     word ptr [ToCnt],0
        jz      TmRet
        dec     word ptr [ToCnt]
TmRet:  popf
        jmp     dword ptr cs:[TmPtr]       ;to standard timer int.

DrvEnd:

;============================================================================
;Laden oder Entfernen des Treibers fr die ser. Schnittstellen
;Startpunkt des .COM-Programms
;Der folgende Programmteil bleibt nach dem Ende des Programms
;nicht im Speicher
;Command-String:  COM2DRV x    ==>  COMx/y-Treiber installieren
;                 COM2DRV x /n ==>  COMx/y-Treiber installieren (n=1,..)
;                 COM2DRV x /u ==>  COMx/y-Treiber wieder entfernen

        ;Bildschirmhinweise

Mess1   db      'installed!',13,10,'$'
Mess2   db      'removed!',13,10,'$'
Mess3   db      'already installed!',13,10,'$'
Mess4   db      'not installed!',13,10,'$'
Mess6   db      13,10,'Calling: COM2DRV ComNr /P<Port,..> /I<IRQ,..> /L<Lines,..>',13,10,'$'

MsgAP   db      4               ;Pos. von ComNr im  nachfolg. String
        db      8               ;Pos. von PortAdr im nachfolg. String
        db      8               ;Pos. von IRQ im nachfolg. String
        db      9               ;Pos. von Lines im nachfolg. String
MessA   db      'COM$: Port $, Int. $, Lines $',13,10,'$'

NLen    equ     4
NBuf    db      NLen dup (?)    ;Puffer fr ASCII-Konv.

assume ds:code

        ;Programmanfang

Init:   mov     ax,cs              ;copy cs to ss
        mov     ss,ax
        mov     sp,offset Tops     ;init stack pointer

        mov     FrstCom,0          ;setze auf COM1
        mov     LastCom,ComAnz-1   ;und COM3
        mov     dx,offset CopRgt
        call    OMess           ;Hinweis auf Copyright ausgeben

        mov     bx,80h             ;Command-String lesen
        mov     ch,0
        mov     cl,[bx]            ;Lnge
        or      cl,cl
        jz      Standrd            ;kein Comm-String

RdPar:  call    GetCh
        jz      Standrd            ;keine Zeichen mehr im Comm-String
TstNxt: cmp     al,'/'             ;Zeichen prfen
        jz      TstOpt             ;Option prfen
        cmp     al,'?'             ;Hilfe
        jz      Error
        call    TstDig             ;Ziffer?
        jc      RdPar              ;nein, nchstes Zeichen
        call    RdNumb             ;COM-Nr.
        dec     dx
        mov     FrstCom,dx
        add     dl,ComAnz-1
        mov     LastCom,dx
        jmp     TstNxt

Standrd:jmp     Install

;-------------------------------------------------------------------------
; Fehlermeldung/Hinweis ausgeben

Error:  mov     dx,offset Mess6
        call    OMess
        jmp     DOS

;-------------------------------------------------------------------------
TstOpt: call    GetCh
        jz      Standrd
        cmp     al,'?'             ;Hilfe
        jz      Error
        or      AL,20H          ;auf Kleinbuchstaben umsetzen
        cmp     al,'u'
        jz      UnInst          ;Treiber entfernen
        cmp     al,'p'
        jz      RdPort          ;Portadressen lesen
        cmp     al,'i'
        jz      RdInt           ;Interrupts lesen
        cmp     al,'l'
        jnz     RdPar

RdLnFi: mov     ch,ComAnz       ;Zeilenfilter lesen
        mov     di,offset LiFiSet

LnFiLp: call    GetCh
        jz      Standrd
        call    TstDig
        jc      Error
        call    RdNumb          ;Wert einlesen
        dec     dx
        jge     LnFi1
        mov     dx,0
LnFi1:  mov     [di],dx
        inc     di
        inc     di
        dec     ch
        jz      NxtArg
        cmp     al,','          ;weitere
        je      LnFiLp
        jmp     Error

RdInt:  mov     ch,ComAnz       ;Interrupts lesen
        mov     di,offset Intrpt
IRQLp:  call    GetCh
        jz      Standrd
        call    TstDig
        jc      Error
        call    RdNumb          ;Interrupt einlesen
        mov     [di],dl
        inc     di
        dec     ch
        jz      NxtArg
        cmp     al,','          ;weitere Int.
        je      IRQLp
        jmp     Error

RdPort: mov     ch,ComAnz
        mov     di,offset Ports
PrtLp:  call    GetCh
        jz      Standrd
        call    TstHex
        jc      Error
        call    RdHex           ;Portadr. einlesen
        mov     [di],dx
        inc     di
        inc     di
        dec     ch
        jz      NxtArg
        cmp     al,','          ;weitere Adr.
        je      PrtLp
        jmp     Error

NxtArg: jmp     TstNxt

;-------------------------------------------------------------------------
UnInst: call    GetInt
        jnz     UnIn1           ;war nicht installiert
        mov     ax,es:[bx+ComDrvID-ComRout]
        cmp     ax,ComDrvID     ;Treiberidentifizierung
        jnz     UnIn1
        jmp     UnIn0           ;Installation entfernen

;-------------------------------------------------------------------------
UnIn1:  mov     dx,offset Mess4
        call    OMess           ;Hinweis ausgeben
        jmp     DOS

;-------------------------------------------------------------------------
Install:mov     si,offset Ports
        mov     di,offset PortAdr    ;Zieltabelle
        mov     cx,ComAnz
        cld
        rep     movsw                ;Ports nach PortAdr kopieren

        mov     cx,ComAnz
        mov     si,offset Intrpt    ;Interrupts setzen
        mov     bx,offset HwInt
HwiLp:  mov     al,[si]         ;Inter. Nr.
        push    cx
        mov     cl,al
        test    al,8            ;2. Contr.?
        jz      Hwi1
        sub     cl,8
Hwi1:   mov     ah,1            ;Maske erzeugen
        shl     ah,cl
        mov     [bx],ax         ;Interrupt Nr. und Maske
        pop     cx
        inc     si
        inc     bx
        inc     bx
        loop    HwiLp

        call    GetInt             ;alte Interrupt-Adr. prfen
        jnz     Inst1              ;sind ungleich ==> installieren

  ;bx =Ptr auf ComRout
        mov     ax,es:[bx+FrstCom-ComRout]       ;erste COM
        mov     cx,es:[bx+LastCom-ComRout]       ;letzte COM
        sub     cx,ax
        inc     cx
        lea     di,[bx+PortAdr-ComRout]
        mov     dx,offset Mess3
        call    OPort
        jmp     DOS                ;sind gleich ==> nichts tun

;-------------------------------------------------------------------------
Inst1:  cli
        mov     word ptr IntPtr,bx   ;alten Interrupt merken
        mov     word ptr IntPtr+2,es
        mov     ax,2500h+BioInt      ;neuen Interrupt setzen
        int     21h

        mov     cx,ComAnz
        xor     si,si
SvHPLp: push    cx
        mov     ax,[si+HwInt]
        mov     ah,35h
        test    al,8                 ;2. Controller?
        jz      SvH1
        add     al,68h
        jmp     short SvH2
SvH1:   add     al,8
SvH2:   int     21h                       ;alten Hardware Int.
        push    si
        shl     si,1
        mov     word ptr [si+HwiPtr],bx   ;merken
        mov     word ptr [si+HwiPtr+2],es
        pop     si
        mov     dx,cs:[si+IntTab]    ;neuen Hardware Interr.
        mov     ah,25h               ;setzen
        int     21h
        pop     cx
        inc     si
        inc     si
        loop    SvHPLp

        mov     word ptr [ToCnt],0   ;reset timer
        mov     ax,3508h             ;get old timer interrupt
        int     21h
        mov     word ptr [TmPtr],bx
        mov     word ptr [TmPtr+2],es
        mov     ah,25h
        mov     dx,offset TmInt      ;set new timer interrupt
        int     21h

        mov     dx,FrstCom
        mov     cx,ComAnz
        call    AllOff
        sti
        mov     dx,offset Mess1
        mov     bx,cs
        mov     es,bx
        mov     ax,FrstCom
        mov     cx,ComAnz
        mov     di,offset PortAdr
        call    OPort             ;Hinweis ausgeben

        mov     dx,offset DrvEnd ;Das Program ist bis "DrvEnd"
                                 ;speicherresident
        int     27h              ;Ende

;-------------------------------------------------------------------------
;alle Coms ausschalten
; dx = 1. ComNr, cx = Anzahl Coms

AllOff: mov     ah,5            ;alle COMs aus
        int     14h
        inc     dx
        loop    AllOff
        ret

;-------------------------------------------------------------------------
UnIn0:  mov     cx,es:[bx+LastCom-ComRout]
        mov     dx,es:[bx+FrstCom-ComRout]         ;Com-Nr.
        sub     cx,dx
        inc     cx                 ;number of COMs
        push    cx
        call    AllOff

        mov     dx,es:[bx+IntPtr-Comrout]
        mov     ds,es:[bx+IntPtr+2-Comrout] ;Original-BIOS-Adresse bei ComRout-4
        mov     ax,2500h+BioInt
        int     21h

        pop     cx              ;Anz. der COMs nach cx

        xor     si,si
HwRLp:  push    si
        mov     ax,es:[si+bx+HwInt-ComRout] ;IRQ Nr. und Maske
        shl     si,1
        mov     dx,es:[si+bx+HwiPtr-ComRout] ;Original-Adresse des Hardw.Int.
        mov     ds,es:[si+bx+HwiPtr-ComRout+2]
        mov     ah,25h
        test    al,8                 ;2. Controller?
        jz      HwR1
        add     al,68h
        jmp     short HwR2
HwR1:   add     al,8
HwR2:   int     21h                  ;alten Hardware Int. setzen
        pop     si
        inc     si
        inc     si
        loop    HwRLp

        mov     dx,es:[bx+TmPtr-ComRout]   ;restore timer interrupt
        mov     ds,es:[bx+TmPtr-ComRout+2]
        mov     ax,2508h
        int     21h

        mov     ax,cs
        mov     ds,ax           ;Setze ds auf cs
        mov     ax,es:[bx+FrstCom-ComRout]       ;erste COM
        mov     cx,es:[bx+LastCom-ComRout]       ;letzte COM
        sub     cx,ax
        inc     cx
        lea     di,[bx+PortAdr-ComRout]
        mov     dx,offset Mess2
        call    OPort           ;Hinweis auf Ports ausgeben

DOS:    mov     ah,0
        int     21h


;-------------------------------------------------------------------------
; COM-Nr. in String (dx) einsetzen und ausgeben
; di: Portadr. / IRQ
; al: COM-Nr.
ComMsg: push    ax
        push    bx
        push    cx
        mov     si,offset MsgAP
        mov     dx,offset MessA
        mov     bl,al
        call    OMess
        mov     al,bl
        call    DConvrt         ;Dez. in ASCII umsetzen
        call    IncDX
        call    OMess
        mov     ax,es:[di]      ;Portadr.
        call    HConvrt         ;Hex. in ASCII umsetzen
        call    IncDX
        call    OMess
        mov     ax,es:[di+2*ComAnz]  ;IRQ
        call    DConvrt
        call    IncDX
        call    OMess
        mov     ax,es:[di+4*ComAnz]  ;Lines
        inc     ax
        call    DConvrt
        call    IncDX
        call    OMess

        pop     cx
        pop     bx
        pop     ax
        ret

IncDX:  mov     al,[si]         ;nchste Pos. im String
        xor     ah,ah
        add     dx,ax
        inc     si
        ret

;-------------------------------------------------------------------------
; Dez-Zahl (al) in ASCII konvertieren und ausgeben
DConvrt:push    dx
        push    bx
        mov     bx,offset NBuf+NLen-1
        mov     byte ptr [bx],'$'
DCvLp:  xor     ah,ah
        mov     dl,10
        div     dl
        add     ah,48           ;ASCII
        dec     bx
        mov     [bx],ah         ;nach Puffer
        or      al,al
        jne     DCvLp
        mov     dx,bx
        call    OMess
        pop     bx
        pop     dx
        ret

;-------------------------------------------------------------------------
; Hex-Zahl (bx) in ASCII konvertieren und ausgeben
HConvrt:push    dx
        mov     bx,offset NBuf+NLen-1
        mov     byte ptr [bx],'$'
HCvLp:  mov     dx,ax
        and     al,0Fh
        add     al,48           ;ASCII
        cmp     al,57
        jbe     HCv1
        add     al,7
HCv1:   dec     bx
        mov     [bx],al         ;nach Puffer
        mov     ax,dx
        mov     cl,4
        shr     ax,cl
        or      al,al
        jne     HCvLp
        mov     dx,bx
        call    OMess
        pop     dx
        ret

;-------------------------------------------------------------------------
; Hinweis auf gewhlte Ports
; ax = erste COM-Nr.-1
; cx = Anzahl COMs
; es:di = Adr. mit Portadr.
OPort:  push    ax
        call    OMess           ;Hinweis ausgeben
        pop     ax
        inc     ax
        xor     ch,ch           ;Zhler
OPLp:   call    ComMsg
        inc     al
        inc     di
        inc     di
        loop    OPLp
        ret

;-------------------------------------------------------------------------
; Bildschirmmeldung
; Pointer auf text in dx

OMess:  mov     ah,9            ;output a string to screen
        int     21h
        ret

;-------------------------------------------------------------------------
; ein Zeichen aus dem Command-String holen
; Zero-Flag = 1 : keine Zeichen mehr, al = 0
;           = 0 : o.k.
GetCh:  xor     al,al
        or      cl,cl           ;keine Zeichen mehr?
        jz      GChRt
        dec     cl
        inc     bx
        mov     al,[bx]
GChRt:  ret

;-------------------------------------------------------------------------
; alte INT14-Adresse holen

GetInt: mov     ax,3500h+BioInt    ;alte Int.-Adr. lesen
        int     21h
        mov     dx,offset ComRout  ;neue INT14-Adresse
        cmp     dx,bx              ;bereits installiert?
        ret

;-------------------------------------------------------------------------
; Hex.Zahl einlesen von Comm. string, 1.Zeichen in al, Pointer in bx
; Wert in dx
RdHex:  xor     ah,ah           ;= 0
        mov     dx,ax
HexLp:  call    GetCh           ;nchstes Zeichen
        jz      HexRet
        call    TstHex
        jc      HexRet          ;keine Ziffer mehr
        call    MulAd16
        jmp     HexLp

HexRet: ret

;-------------------------------------------------------------------------
; Zeichen in al auf Hex-Ziffer prfen und in Binr wandeln
; Carry Flag = 1: keine Hex-Ziffer
;            = 0: ok

TstHex: cmp     al,30h          ;Ziffer
        jb      TstHexN         ;no
        cmp     al,39h
        ja      TstHexA         ;no
        sub     al,30H          ;convert to binary
        jmp     TstHexR

TstHexA:cmp     al,'A'          ;A..F
        jb      TstHexN         ;no
        cmp     al,'F'
        ja      TstHexB
        sub     al,37h          ;convert to binary
        jmp     TstHexR

tstHexB:cmp     al,'a'          ;a..f
        jb      TstHexN         ;no
        cmp     al,'f'
        ja      TstHexN
        sub     al,57h
TstHexR:clc                     ;clear carry flag
        ret

TstHexN:stc                     ;set carry flag
        ret

;-------------------------------------------------------------------------
; Dez.Zahl einlesen von Comm. string, 1.Zeichen in al, Pointer in bx
; Wert in dx

RdNumb: xor     ah,ah           ;= 0
        mov     dx,ax
DigLp:  call    GetCh           ;nchstes Zeichen
        jz      DigRet
        call    TstDig
        jc      DigRet          ;keine Ziffer mehr
        call    MulAd10
        jmp     DigLp
DigRet: ret

;-------------------------------------------------------------------------
; Zeichen in al auf Ziffer prfen und in BCD wandeln
; Carry Flag = 1: keine Hex-Ziffer
;            = 0: ok

TstDig: cmp     al,30h          ;Ziffer
        jb      TstDigN         ;no
        cmp     al,39h
        ja      TstDigN         ;no
        sub     al,30H          ;convert to binary and reset zero flag
        clc                     ;clear carry flag
        ret

TstDigN:stc                     ;set carry flag
        ret

;-------------------------------------------------------------------------
; Multipliziere dx mit 10 und addiere ax

MulAd10:shl     dx,1            ;*2
        push    dx
        shl     dx,1            ;*4
        shl     dx,1            ;*8
        add     dx,ax
        pop     ax              ;dx*2
        add     dx,ax
        ret

;-------------------------------------------------------------------------
; Multipliziere dx mit 16 und addiere ax

MulAd16:push    cx
        mov     cl,4
        shl     dx,cl           ;*16
        pop     cx
        add     dx,ax
        ret

;-------------------------------------------------------------------------
; Standard-Portadressen:

Ports   dw      3F8H,2F8H  ;COM1-2-Port
Intrpt  db      4,3        ;COM1-2-Int.

IntTab  dw      Com1Int,Com2Int
;-------------------------------------------------------------------------
;space for user stack
        dw      100h dup (?)
Tops    label word

code    ends
        end     start
