        page    60, 132
        title   Driver for COM1,2 with Int. 4 or 3 (Vers. 3.3) In/Out

;-------------------------------------------------------------------------;  Vers. 3: Interruptuntersttzung fr Senden und Empfangen
;  J. Rathlev, Jun. 1995
;  Vers. 3.2 mit Interrupt-Pufferung der Ein- und Ausgabe (95)
;  Vers. 3.3 mit Zeilenzhlung bei <CR>
;--------------------------------------------------------------------------
;  Das nachfolgende Programm emuliert den BIOS-Interrupt 14
;  (comm. port) unter  Benutzung eines interruptgesteuerten
;  Buffers fr den Datenempfang und das Datensenden
;  des Kanales COMx. (x=1 oder 2)
;  Die alten Funktionen
;               ah = 0 : Initialisierung (Baud Rate etc. und siehe Fkt. 4)
;                  = 1 : Daten senden
;                  = 2 : Daten empfangen
;                  = 3 : Status lesen
;  sind erhalten geblieben.
;  Hinzu ist gekommen
;               ah = 4 : Modem einschalten, Buffer Init. und Interrupt an
;                        (auch in Funktion 0 enthalten)
;                  = 5 : Modem ausschalten, Interrupt aus
;                  = 6 : Line Control Register auslesen
;                  = 7 : Empf.-/Sendebuffer Counter auslesen
;                  = 8 : Treiber identifizierung
;
;
;  Das Programm wird nach dem Laden des Betriebssystem MS-DOS einmal
;  aufgerufen (z.B. mit "autoexec.bat") und bleibt dann im Speicher
;  resident:
;  Command-String:  COMDRV    ==>  COM-Treiber installieren (Standard)
;                   COMDRV /n ==>  COM-Treiber installieren (n=1-4)
;                                   mit alternativen Portadressen
;                                   (siehe Tabellen
;                                   am Ende des Programms)
;                   COMDRV /u ==>  COM-Treiber wieder entfernen
;
;  Alle Aufrufe von Int 14 durch irgendein Programm werden
;  abgefangen und durch die unten beschriebenen Routinen bedient.
;
;  Hardware-Voraussetzungen:
;       1. Interface fr COM1 und/oder COM2
;       2. Die Jumper auf der Platine mssen auf Interrupt 4 oder 3
;          eingesetzt sein.
;       3. Die Schalterstellung fr die Portadressen mu mit der Auswahl von
;          "n" (siehe oben) bereinstimmen:
;           n = 1 ==> COM1 (3F8) mit IRQ 4 (Standardeinstellung fr COM1)
;               2 ==> COM2 (2F8) mit IRQ 3 (Standardeinstellung fr COM2)
;               3 ==> COM1 (3F8) mit IRQ 3
;               4 ==> COM2 (2F8) mit IRQ 4
;               5 ==> COM1 (2F8) mit IRQ 4
;               6 ==> COM2 (3F8) mit IRQ 3
;               7 ==> COM1 (2F8) mit IRQ 3
;               8 ==> COM2 (3F8) mit IRQ 4
;
;  Schnittstellen-Protokoll:
;    Senden: Die Funktion 0 oder 4 (Modem an) gibt die
;       Signale DTR (pin 20) und RTS (pin 4) aus (+ 12 V).
;       Vor dem Senden eines Zeichens werden DSR (pin 6) und CTS (pin 5)
;       abgefragt. Sie mssen beide gesetzt sein (+12 V). Wenn diese
;       Bedingung nicht erfllt ist, erfolgt ein Rcksprung aus der
;       Senderoutine (Funktion 1) mit dem Time-out-bit (ah-bit 7)
;       auf 1 gesetzt.
;    Empfangen: Es ist grundstzlich DTR (pin 20) gesetzt. Alle ankommenden
;       Zeichen werden im Buffer zwischengespeichert (sofern mit Funktion 0
;       oder 4 das Modem eingeschaltet wurde).
;       Der Buffer kann max. 256 Zeichen pro Schnittstelle mit dem
;       zugehrigen Fehlerstatus (Overrun-, Parity- oder Framing-Error)
;       aufnehmen. Weitere Zeichen werden ignoriert.
;
;  Status-Abfrage:
;       Die Funktion 3 liest den gleichen Status wie im Original-
;       BIOS mit Ausnahme des DR-Bits (Data Ready, ah-bit 0). Dieses Bit
;       zeigt hier an, da im Input-Buffer mindestens ein Zeichen gespeichert
;       ist, das mit der Funktion 2 abgeholt werden kann.
;
;-------------------------------------------------------------------------


code    segment
        assume  cs:code
        assume  ds:code
        assume  es:nothing
        assume  ss:nothing


; BIOS Interrupt Number

BioInt  equ     14h

CR      equ     0DH     ;carriage return

; Registers of 8250:

IER     equ     1       ; Interrupt Enable Register
IIR     equ     2       ; Interrupt Identification Register
LCR     equ     3       ; Line Control Register
MCR     equ     4       ; MODEM Control Register
LSR     equ     5       ; Line Status Register
MSR     equ     6       ; MODEM Status Register

EOI     equ     20h     ; end of interrupt

;-------------------------------------------------------------------------
        org     100h

start:  jmp     init

DrvName : db 'COMDRV.COM (Vers. 3.3)',0
CopRgt  : db 'PC-COM-Driver (Vers. 3.31 - (c) J.Rathlev 1995) $'

; input buffer

RcvLen    equ     256            ;receiver buffer length
SndLen    equ     512            ;send buffer length

RcvBuf  dw      RcvLen dup (?)   ;receive buffer
SndBuf  db      SndLen dup (?)   ;send buffer

; bufferpointer and counter

;receive buffer
RcvBPtr dw      RcvBuf
RcvCnt  dw      ?       ;buffer count
LineCnt dw      ?       ;input buffer line counter
RcvIPnt dw      ?       ;input buffer pointer
RcvOPnt dw      ?       ;output buffer pointer

;send buffer
SndBPtr dw      SndBuf
SndCnt  dw      ?       ;buffer count
SndIPnt dw      ?       ;input buffer pointer
SndOPnt dw      ?       ;output buffer pointer

; interrupt masks and pointer

PortAdr   dw      ?       ;RS232 port Adr.
HwInt     dw      ?       ;IRQ Number and Mask

HwiPtr    dd      ?       ;Pointer to old Hardware Int.
ComNr     db      ?       ;Number of COM
ComMd     db      ?       ;Mode
IntPtr    dd      ?       ;Pointer to BIOS-INT14
ComDrvID  dw      0F133h               ;driver identifier word

;-----------------------------------------------------------------
; the following routine "ComRout" is the new INT14 for COMx
; port number in dx
; function in ah:
;-------------------------------------------------------------------------
;  0    initialize the comm. port
;       enable interrupts from line 4
;       al contains parameter for init.:
;       normal: (bit 1 = 1)
;       bit
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -    110         X0 - none    0 - 1   10 - 7 bits
;       001 -    150         01 - odd     1 - 2   11 - 8 bits
;       010 -    300         11 - even
;       011 -    600
;       100 -   1200
;       101 -   2400
;       110 -   4800
;       111 -   9600
;       on return status in ah,al as function 3
;
;       extended: (bit 1 = 0)
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -     50         X0 - none    0 - 1   00 - 7 bits
;       001 -     75         01 - odd     1 - 2   01 - 8 bits
;       010 -    134.5       11 - even
;       011 -   1800
;       100 -  19200
;       101 -  38400
;       110 -  57600
;       111 - 115200
;
;       on return status in ah,al as function 3
;
;-------------------------------------------------------------------------
;  1    send a char. in al to comm. line
;       al is saved
;       return with
;          ah bit 7 = 1 : unable to send char. to line
;                   = 0 : bit 0-6 contain line status
;
;-------------------------------------------------------------------------
;  2    read a char. to al from comm. line via buffer
;       return with
;          ah bit 7 = 1 : no char. in buffer
;                   = 0 : bit 1,2,3,4 contain line status
;
;-------------------------------------------------------------------------
;  3    read status from comm. port
;       return with
;          ah bit 7 : always 0
;                 6 : transmitter shift reg. empty
;                 5 : transmitter holding reg. empty
;                 4 : break interrupt
;                 3 : framing error
;                 2 : parity error
;                 1 : overrun error
;                 0 : data received (in buffer)
;          al bit 7 : received line signal detect
;                 6 : ring indicator
;                 5 : data set ready
;                 4 : clear to send
;                 3 : delta received line signal detect
;                 2 : trailing edge ring indicator
;                 1 : delta data set ready
;                 0 : delta clear to send
;
;-------------------------------------------------------------------------
;  4    switch modem on (set DTR and RTS, enable interrupts)
;-------------------------------------------------------------------------
;  5    switch modem off (reset DTR and RTS, interrupt disable)
;-------------------------------------------------------------------------
;  6    read line control register to al
;-------------------------------------------------------------------------
;  7    read buffer counter to ax, buffer length to dx
;-------------------------------------------------------------------------
;  8    driver identifikation: set al=0 on call
;                              single COM-driver installed:
;                                     return al=$F1
;                              otherwise al unchanged
;-------------------------------------------------------------------------

ComRout:sti
        push    ds
        push    si
        push    di
        mov     si,cs           ;copy cs to ds
        mov     ds,si
        cmp     dl,ComNr        ;number of emulated COM
        jne     Direct          ;above, jump to BIOS INT14
        push    bx
        push    cx
        push    dx
        mov     bl,ah           ;function to bl
        xor     bh,bh
        shl     bx,1            ;*2
        cmp     bx,FctNr        ;legal function?
        jnc     ComFin          ;no, return imm.
        mov     dx,[PortAdr]    ;Portaddr
        jmp     word ptr [bx+offset FctTab]   ;branch

FctTab  dw      ComSet          ;set baud rate and format
        dw      ComXmit         ;send char. via ser. port
        dw      ComRecv         ;get a received char. from buffer
        dw      ComStat         ;get modem status
        dw      ComOn           ;switch modem on
        dw      ComOff          ;switch modem off
        dw      ComLCR          ;read LCR
        dw      ComBCnt         ;read buffer counter
        dw      ComID           ;driver identifier
FctNr   equ     $-FctTab

;-------------------------------------------------------------------------
ComFin: pop     dx
        pop     cx
        pop     bx
        jmp     ComRet


;jump to BIOS-INT14 routine

Direct: cmp     ah,3
        jbe     ClBios  ;Original BIOS: function 0 - 3
        xor     ax,ax
        xor     dx,dx
        jmp     ComRet
ClBios: pushf           ;prepare to emulate int-call
        cli
        call    dword ptr [IntPtr]    ;call to BIOS
ComRet: pop     di
        pop     si
        pop     ds
        iret


;-------------------------------------------------------------------------
;baud rate table

BTab1   dw      1047            ;110 baud
        dw      768             ;150 baud
        dw      384             ;300 baud
        dw      192             ;600 baud
        dw      96              ;1200 baud
        dw      48              ;2400 baud
        dw      24              ;4800 baud
        dw      12              ;9600 baud
BTab2   dw      2304            ;50 baud
        dw      1536            ;75 baud
        dw      857             ;134.5 baud
        dw      64              ;1800 baud
        dw      6               ;19200 baud
        dw      3               ;38400 baud
        dw      2               ;57600 baud
        dw      1               ;115200 baud

;-------------------------------------------------------------------------
;set ser. port to new baud rate and format in ax
;dx = port addr., ds = cs
;new values in al

ComSet: push    ax              ;save
        call    ComRlse
        pop     ax
        push    dx              ;save port addr.
        mov     ah,al           ;save
        add     dx,LCR          ;line control reg.
        mov     al,80H          ;set DLAB
        out     dx,al
        mov     bl,ah
        mov     cl,4            ;rot count
        rol     bl,cl           ;rotate 4 bits
        and     bx,0EH          ;mask out baud rate
        test    ah,2            ;extended mode?
        jz      ComExt
        add     bx,offset BTab1 ;pointer to 1. divisor table
        jmp     short ComNor
ComExt: add     bx,offset BTab2 ;pointer to 2. divisor table
ComNor: dec     dx
        dec     dx              ;divisor latch (DLM)
        mov     al,[bx+1]
        out     dx,al
        dec     dx              ;divisor latch (DLL)
        mov     al,[bx]
        out     dx,al
        add     dx,LCR          ;line control reg.
        mov     al,ah           ;format bits
        or      al,2            ;set bit 1
        and     al,1FH          ;mask out
        out     dx,al
        pop     dx              ;restore port base addr.
        call    ComInit         ;initialize
        jmp     ComStat


;-------------------------------------------------------------------------
;wait until char. received and read char. from buffer
;dx = port addr., ds = cs
;return with char. in al, ah = 80h on time out, else = Line Status

ComRecv:lea     bx,[RcvCnt]    ;test buffer count
        xor     cx,cx
ComRWt: cmp     word ptr [bx],0
        jnz     ComRc2          ;char. received
        loop    ComRWt          ;wait until not zero
        mov     ah,80H          ;time out
        jmp     ComFin          ;return imm.

ComRc2: dec     word ptr [bx]   ;count one down
        mov     bx,[RcvOPnt]    ;buffer pointer
        mov     di,[RcvBPtr]    ;buffer base addr.
        mov     ax,[bx+di]      ;get status and char. from buffer
        inc     bx              ;incr. buffer pointer
        inc     bx
        cmp     bx,2*RcvLen     ;buffer length
        jc      ComRc1
        xor     bx,bx           ;set to first buffer addr.
ComRc1: mov     [RcvOPnt],bx   ;and store
        cmp     al,CR           ;carriage return?
        jnz     ComRc3
        cmp     word ptr [LineCnt],0
        jz      ComRc3
        dec     word ptr [LineCnt]
ComRc3: jmp     ComFin


;-------------------------------------------------------------------------
;send char. in al via buffer
;dx = port addr., ds = cs
;return with ah (bit 7) = 1 buffer full, al = char
;            ah (bit 7) = 0 on success, al =char

ComXmit:lea     bx,[SndCnt]    ;address of buffer counter
        cmp     word ptr [bx],SndLen
        jge     ComXm1            ;buffer full

        cli
        or      word ptr [bx],0
        jnz     ComXm3            ;interrupt active
        mov     ah,al             ;save char.
        add     dx,IER
        in      al,dx
        or      al,2              ;enable THRE-Interr.
        out     dx,al
        mov     al,ah
ComXm3: inc     word ptr [bx]     ;incr. buffer count
        mov     bx,[SndIPnt]   ;buffer pointer
        mov     di,[SndBPtr]   ;buffer base addr.
        mov     [bx+di],al        ;store to buffer
        inc     bx                ;incr. buffer pointer
        cmp     bx,SndLen         ;buffer length
        jc      ComXm2
        xor     bx,bx             ;set to first buffer addr.
ComXm2: mov     [SndIPnt],bx   ;and store
        sti
        xor     ah,ah
        jmp     comfin

ComXm1: mov     ah,80H          ;set error bit
        jmp     ComFin


;-------------------------------------------------------------------------
;read status from ser. port
;dx = port addr., ds = cs
;ah and al contain the status bits like the orig. BIOS-Int.
;except bit 0 in ah (data ready). This bit is set if the input buffer
;is not empty (RcvCnt)

ComStat:add     dx,LSR          ;line status reg.
        in      al,dx
        mov     ah,al
        inc     dx              ;modem status reg.
        in      al,dx
        and     ah,0FEH         ;mask out bit 0
        mov     bx,[RcvCnt]
        or      bx,bx
        jz      ComSt1
        or      ah,1            ;set bit 0
ComSt1: jmp     ComFin


;-------------------------------------------------------------------------
;switch modem on
;dx = port addr., ds = cs

ComOn:  call    ComInit
        jmp     ComFin


;-------------------------------------------------------------------------
;switch modem off
;dx = port addr., d     s = cs

ComOff: call    ComRlse
        jmp     ComFin


;-------------------------------------------------------------------------
; read line control register
; dx = port addr., ds = cs

ComLCR: add     dx,LCR
        in      al,dx
        jmp     ComFin


;-------------------------------------------------------------------------
; read buffer counter
; dx = port addr., ds = cs
; al = 0 (Rcv)
;      1 (Snd)
; return: ax = buffer counter
;         cx = line counter (on receive mode)
;         dx = buffer length

ComBCnt:pop     dx
        pop     cx
        pop     bx
        or      al,al
        jz      GetRcv
        mov     ax,[SndCnt]
        mov     dx,SndLen
        jmp     ComRet

GetRcv: mov     ax,[RcvCnt]
        mov     cx,[LineCnt]
        mov     dx,RcvLen
        jmp     ComRet

;-------------------------------------------------------------------------
; get driver identification
; al = $F1 for single COM-driver
; al = version
; cl = IRQ
; dx = port
; es:bx = pointer to driver name

ComID:  mov     ax,dx           ;save port addr.
        pop     dx
        pop     cx
        pop     bx
        lea     bx,[DrvName]
        mov     dx,ax
        mov     ax,cs
        mov     es,ax
        mov     cx,[HwInt]      ;interrupt
        mov     ax,ComDrvID     ;identification byte
        jmp     ComRet

;-------------------------------------------------------------------------
;subroutine test ser. port status
;in dx port addr.of status reg., in bh status bits to test
;destroyed: cx
;return with zero flag = 0 if time out, status in ah

TstStat:xor     cx,cx           ;clear
TstSta1:in      al,dx
        mov     ah,al
        xor     al,0FFH         ;complement
        test    al,bl           ;test bits
        jz      TstStaR         ;o.k. return
        loop    TstSta1         ;try again
        or      bh,bh           ;clear zero flag
TstStaR:ret


;-------------------------------------------------------------------------
;subroutine init. buffer and set modem on
;dx = base port addr.; ds = cs

ComInit:push    dx              ;save port-addr.
        xor     ax,ax
        cli
        mov     [RcvCnt],ax    ;receive buffer count
        mov     [LineCnt],ax   ;         line counter
        mov     [RcvIPnt],ax   ;        buffer pointer
        mov     [RcvOPnt],ax
        mov     [SndCnt],ax    ;send buffer count
        mov     [SndIPnt],ax   ;     buffer pointer
        mov     [SndOPnt],ax
        sti
        in      al,dx           ;clear receiver buffer reg.
        add     dx,MCR          ;modem control reg. - modem on
        mov     al,0BH          ;set DTR, RTS and OUT2
        out     dx,al
        sub     dx,MCR-IER      ;enable interrupt
        mov     al,1            ;for receive
        out     dx,al
        mov     ax,[HwInt]      ;set interrupt controller
        not     ah
        in      al,21h          ;interrupt mask
        and     al,ah           ;enable interrupt
        out     21h,al
        pop     dx
        ret


;-------------------------------------------------------------------------
;subroutine release and modem off
;dx = base port addr.; ds = cs

ComRlse:
        push    dx              ;save port-addr.
        mov     ax,[HwInt]
        in      al,21h          ;interrupt mask
        or      al,ah           ;disable interrupt
        out     21h,al
        inc     dx
        xor     al,al
        out     dx,al           ;IER
        add     dx,3            ;MCR - set modem off
        out     dx,al
        pop     dx
        ret


;-------------------------------------------------------------------------
;receive char. interrupt
; si =  2 * rel. ComNr
; dx =  portaddr. + IIR

assume ds:code

RcvInt: add     dx,LSR-IIR      ;Line status reg.
        in      al,dx
        and     al,1EH          ;mask out error bits
        mov     ah,al
        sub     dx,LSR
        in      al,dx           ;read char. from ser.port

        lea     bx,[RcvCnt]            ;address of buffer counter
        cmp     word ptr [bx],RcvLen
        jz      RcvFul                    ;buffer full

        cmp     al,CR                     ;carriage return
        jnz     RcvI1
        inc     word ptr [LineCnt]     ;incr. line count
RcvI1:  inc     word ptr [bx]             ;incr. buffer count
        mov     bx,[RcvIPnt]           ;buffer pointer
        mov     di,[RcvBPtr]           ;buffer base addr.
        mov     [bx+di],ax                ;store to buffer
        inc     bx                        ;incr. pointer
        inc     bx
        cmp     bx,2*RcvLen
        jc      StBiP
        xor     bx,bx                     ;init. buffer addr.
StBiP:  mov     [RcvIPnt],bx           ;store

RcvFul: add     dx,IIR
        ret

;send char. interrupt

SndInt: or      word ptr [SndCnt],0
        jz      SndOff                  ;buffer not busy
        dec     word ptr [SndCnt]           ;decr. buffer count
        sub     dx,IIR
        mov     bx,[SndOPnt]         ;buffer pointer
        mov     di,[SndBPtr]         ;buffer base addr.
        mov     al,[bx+di]              ;load from buffer
        out     dx,al                   ;send char.
        inc     bx                      ;incr. pointer
        cmp     bx,SndLen
        jc      StBoP
        xor     bx,bx                   ;init. buffer addr.
StBoP:  mov     [SndOPnt],bx         ;store
        add     dx,IIR
        ret

SndOff: dec     dx              ;to IER
        in      al,dx
        and     al,0Dh          ;disable THRE-Interr.
        out     dx,al
        inc     dx              ;to IIR
        ret


; MSR interrupt, should not happen

MSRInt: add     dx,MSR-IIR
        in      al,dx
        sub     dx,MSR-IIR
        ret

; LSR interrupt, should not happen

LSRInt: add     dx,LSR-IIR
        in      al,dx
        sub     dx,LSR-IIR
        ret

assume ds:nothing

;-------------------------------------------------------------------------
;interrupt routine for ser. port (int 0CH/0BH)

ComInt: sti
        push    ax
        push    bx
        push    cx
        push    dx
        push    di
        push    ds
        mov     ax,cs
        mov     ds,ax
        mov     cx,5            ;try 5 times

assume ds:code

IntLp:  mov     dx,[PortAdr]
        add     dx,IIR
IPoll:  in      al,dx           ;read IIR
        test    al,1            ;interrupt pending
        jnz     IntEnd          ;no
        and     ax,06h
        mov     bx,ax
        call    word ptr [bx+offset IntBra]
        jmp     short IPoll

assume ds:nothing

IntEnd: loop    IntLp           ;try again

        pop     ds
        pop     di
        pop     dx
        pop     cx
        pop     bx
        mov     ax,[HwInt]
        cli
        test    al,8
        mov     al,EOI          ;end of  interrupt
        jz      EOInt1          ;1. contr.
        out     0A0h,al         ;end of  interrupt (contr.2)
        jcxz    EOInt1          ;wait state
EOInt1: out     20h,al          ;end of  interrupt (contr.1)
        pop     ax
        iret

IntBra: dw      MSRInt
        dw      SndInt
        dw      RcvInt
        dw      LSRInt

DrvEnd:

;============================================================================
;Laden oder Entfernen des Treibers fr die ser. Schnittstellen
;Startpunkt des .COM-Programms
;Der folgende Programmteil bleibt nach dem Ende des Programms
;nicht im Speicher
;Command-String:  COMDRV    ==>  COM-Treiber installieren
;                 COMDRV /n ==>  COM-Treiber installieren (n=1,..)
;                 COMDRV /u ==>  COM-treiber wieder entfernen

        ;Bildschirmhinweise

Mess1   db      'installed!',13,10,'$'
Mess2   db      'removed!',13,10,'$'
Mess3   db      'already installed!',13,10,'$'
Mess4   db      'not installed!',13,10,'$'
Mess5   db      'Mode$'
Mess6   db      13,10,'Calling: COMDRV /Mode',13,10,'possible Modes:',13,10,'$'
MMd1    db      ' 1: COM1: Port 3F8, Int. 4',13,10,'$'
MMd2    db      ' 2: COM2: Port 2F8, Int. 3',13,10,'$'
MMd3    db      ' 3: COM1: Port 3F8, Int. 3',13,10,'$'
MMd4    db      ' 4: COM2: Port 2F8, Int. 4',13,10,'$'
MMd5    db      ' 5: COM1: Port 2F8, Int. 4',13,10,'$'
MMd6    db      ' 6: COM2: Port 3F8, Int. 3',13,10,'$'
MMd7    db      ' 7: COM1: Port 2F8, Int. 3',13,10,'$'
MMd8    db      ' 8: COM2: Port 3F8, Int. 4',13,10,'$'

MMTab   dw      MMd1,MMd2,MMd3,MMd4,MMd5,MMd6,MMd7,MMd8

        ;Programmanfang

Init:   mov     ax,cs              ;copy cs to ss
        mov     ss,ax
        mov     sp,offset Tops     ;init stack pointer

        mov     dx,offset CopRgt
        call    OMess           ;Hinweis auf Copyright ausgeben

        mov     bx,80h             ;Command-String lesen
        mov     ch,0
        mov     cl,[bx]            ;Lnge
        or      cl,cl
        jz      Standrd            ;kein Comm-String

RdPar:  inc     bx
        mov     al,[bx]
        cmp     al,'/'             ;Zeichen prfen
        jz      TstOpt             ;Option prfen
        cmp     al,'?'             ;Hilfe
        jz      Error
        loop    RdPar

Standrd:mov     dx,1               ;setze auf Mode 1
        jmp     Install

;-------------------------------------------------------------------------
; Fehlermeldung/Hinweis ausgeben

Error:  mov     dx,offset Mess6
        call    OMess
        mov     cx,ModeAnz
        mov     bx,offset MMTab
ErrLp:  mov     dx,[bx]
        call    OMess
        inc     bx
        inc     bx
        loop    ErrLp
        jmp     DOS

;-------------------------------------------------------------------------
TstOpt: dec     cx
        mov     al,[bx+1]       ;nchstes Zeichen prfen
        cmp     al,'u'
        jz      UnInst          ;Treiber entfernen
        cmp     al,'U'
        jz      UnInst          ;dto.
        call    TstDig          ;Ziffer?
        jz      RdPar           ;nein, nchstes Zeichen
        xor     ah,ah           ;= 0
        mov     dx,ax
        dec     cx
DigLp:  inc     bx
        mov     al,[bx+1]       ;nchstes Zeichen
        call    TstDig
        jz      Install         ;keine Ziffer mehr
        call    MulAd10
        loop    DigLp

Install:cmp     dx,0
        jbe     Error           ;falscher Mode
        cmp     dx,ModeAnz      ;max. Anzahl von Modes
        ja      Error           ;falscher Mode
        mov     ComMd,dl        ;Mode speichern
        push    dx              ;Mode in dl
        shl     dx,1            ;*2
        mov     si,dx
        mov     si,cs:[si+ModeTab-2] ;Tabelle mit Portadressen, etc.
        mov     ax,[si]
        mov     PortAdr,ax         ;store port address
        inc     si
        inc     si
        mov     ax,[si]
        mov     cx,ax
        mov     ah,1
        shl     ah,cl
        mov     HwInt,ax           ;Interrupt Nr. und Maske
        inc     si
        inc     si
        mov     ax,[si]            ;emulated COM-Nr.
        mov     ComNr,al

        call    GetInt             ;alte Interrupt-Adr. prfen
        jnz     Inst1              ;sind ungleich ==> installieren
        mov     dx,offset Mess3
        pop     bx                 ;Mode
        call    OMode
        jmp     DOS                ;sind gleich ==> nichts tun

;-------------------------------------------------------------------------
UnInst: call    GetInt
        jnz     UnIn1           ;war nicht installiert
        mov     ax,es:[bx+ComDrvID-ComRout]
        cmp     ax,ComDrvID     ;Treiberidentifizierung
        jnz     UnIn1
        jmp     UnIn0           ;Installation entfernen

;-------------------------------------------------------------------------
Inst1:  mov     word ptr IntPtr,bx   ;alten Interrupt merken
        mov     word ptr IntPtr+2,es
        mov     ax,2500h+BioInt      ;neuen Interrupt setzen
        int     21h

        mov     ax,HwInt
        mov     ah,35h
        add     al,8
        int     21h                    ;alten Hardware Int.
        mov     word ptr [HwiPtr],bx   ;merken
        mov     word ptr [HwiPtr+2],es

        mov     dx,cs:[IntTab]    ;neuen Hardware Interr.
        mov     ah,25h            ;setzen
        int     21h

        mov     dl,ComNr
        xor     dh,dh
        mov     ah,5            ;Modem aus
        int     14h

        mov     dx,offset Mess1
        pop     bx                 ;Mode
        call    OMode

        mov     dx,offset DrvEnd ;Das Program ist bis "DrvEnd"
                                 ;speicherresident
        int     27h              ;Ende


;-------------------------------------------------------------------------
UnIn1:  mov     dx,offset Mess4
        call    OMess           ;Hinweis ausgeben
        jmp     DOS


UnIn0:  mov     dx,es:[bx+ComNr-ComRout]    ;Mode und COM-Nr.
        push    dx              ;Mode retten
        xor     dh,dh           ;ComNr in dx
        mov     ah,5            ;Modem aus
        int     14h

        mov     dx,es:[bx+IntPtr-ComRout]
        mov     ds,es:[bx+IntPtr+2-ComRout]       ;Original-BIOS-Adresse bei ComRout-4
        mov     ax,2500h+BioInt
        int     21h

        mov     dx,es:[bx+HwiPtr-ComRout]
        mov     ds,es:[bx+HwiPtr+2-ComRout]       ;Original-Adresse des Hardw.Int.
        mov     ax,es:[bx+HwInt-ComRout]
        mov     ah,25h
        add     al,8
        int     21h

        mov     dx,offset Mess2
        mov     bx,cs
        mov     ds,bx
        pop     bx                 ;Mode
        mov     bl,bh
        xor     bh,bh
        call    OMode

DOS:    mov     ah,0
        int     21h


;-------------------------------------------------------------------------
; Hinweis auf gewhlten Mode

OMode:  call    OMess           ;Hinweis ausgeben
        mov     dx,offset Mess5
        call    OMess
        shl     bx,1            ;Mode *2
        mov     dx,[bx+MMTab-2]
        call    OMess
        ret

;-------------------------------------------------------------------------
; Bildschirmmeldung
; Pointer auf text in dx

OMess:  mov     ah,9            ;output a string to screen
        int     21h
        ret

;-------------------------------------------------------------------------
; alte INT14-Adresse holen

GetInt: mov     ax,3500h+BioInt    ;alte Int.-Adr. lesen
        int     21h
        mov     dx,offset ComRout  ;neue INT14-Adresse
        cmp     dx,bx              ;bereits installiert?
        ret

;-------------------------------------------------------------------------
; Zeichen in al auf Ziffer prfen
; Zero Flag = 1: nein
;           = 0: ja

TstDig: cmp     al,30h          ;Ziffer
        jb      TstDigN         ;no
        cmp     al,39h
        ja      TstDigN         ;no
        sub     al,30H          ;convert to binary and reset zero flag
        ret

TstDigN:xor     al,al           ;set zero flag
        ret

;-------------------------------------------------------------------------
; Multipliziere dx mit 10 und addiere ax

MulAd10:shl     dx,1            ;*2
        push    dx
        shl     dx,1            ;*4
        shl     dx,1            ;*8
        add     dx,ax
        pop     ax              ;dx*2
        add     dx,ax
        ret

;-------------------------------------------------------------------------
; Portadressen:

ModeTab dw      Mode1,Mode2,Mode3,Mode4,Mode5,Mode6,Mode7,Mode8
ModeAnz equ     ($-ModeTab)/2

Mode1   dw      3F8H,4,0   ;COM-Port, COM-Int.
Mode2   dw      2F8H,3,1
Mode3   dw      3F8H,3,0
Mode4   dw      2F8H,4,1
Mode5   dw      2F8H,4,0
Mode6   dw      3F8H,3,1
Mode7   dw      2F8H,3,0
Mode8   dw      3F8H,4,1

IntTab  dw      ComInt

;-------------------------------------------------------------------------
;space for user stack
        dw      100h dup (?)
Tops    label word

code    ends
        end     start

