     page    60, 132
     title   PC32 driver - Vers. 3.31

;-------------------------------------------------------------------------
;  J. Rathlev, Jun. 1995
;  Vers. 3.2 mit Interrupt-Pufferung fr Ein- und Ausgabe
;            IRQ  : 3,4,5,9
;  Vers. 3.3 mit Zeilenzhlung bei <CR>
;-------------------------------------------------------------------------
;  Das nachfolgende Programm emuliert den BIOS-Interrupt 14
;  (comm. port) unter  Benutzung eines interruptgesteuerten
;  Buffers fr Datenempfang und -senden fr mehrere Kanle (hier COMn-(n+7).
;  Die alten Funktionen
;               ah = 0 : Initialisierung (Baud Rate etc. und siehe Fkt. 4)
;                  = 1 : Daten senden
;                  = 2 : Daten empfangen
;                  = 3 : Status lesen
;  sind erhalten geblieben.
;  Hinzu ist gekommen
;               ah = 4 : Modem einschalten, Buffer Init. und Interrupt an
;                        (auch in Funktion 0 enthalten)
;                  = 5 : Modem ausschalten, Interrupt aus
;                  = 6 : Line Control Register auslesen
;                  = 7 : Empf.-/Sendebuffer Counter auslesen
;                  = 8 : Treiber Identifizierung
;
;  Das Programm wird nach dem Laden des Betriebssystem MS-DOS einmal
;  aufgerufen (z.B. mit "autoexec.bat") und bleibt dann im Speicher
;  resident:
;  Command-String:  PC32DRVN    ==>  PC32-Treiber installieren (Standard)
;                                      (siehe unten n = 1)
;                   PC32DRVN /n ==>  PC32-Treiber installieren (n=1..35)
;                                      mit alternativen Portadressen
;                                      (siehe PC32-Handbuch oder Tabelle
;                                       weiter unten)
;                   PC32DRVN /u ==>  PC32-Treiber wieder entfernen
;
;  Alle Aufrufe von Int 14 durch irgendein Programm werden
;  abgefangen und durch die unten beschriebenen Routinen bedient.
;
;  Bei Benutzung der Turbo-Pascal-Unit "TpComX" werden auch die Funktionen
;  4 bis 8 untersttzt.
;
;  Hardware-Voraussetzungen:
;       1. Meilhaus PC31/32 Platine.
;       2. Die Jumper auf der Platine mssen alle auf den gleichen
;          Interrupt eingestellt werden:
;          Standard: Interrupt 4 oder 3
;                    (durch Umlten auch 5 oder 9)
;       3. Die Schalterstellung fr die Portadressen mu mit der Auswahl von
;          "n" (siehe weiter unten) bereinstimmen:
;           n = 1 ==> OFF ON  ON  (Standardeinstellung) mit IRQ 4
;               2 ==> ON  OFF OFF                       mit IRQ 4
;               3 ==> ON  OFF ON                        mit IRQ 4
;               4 ==> ON  ON  OFF                       mit IRQ 4
;               5 ==> ON  ON  ON                        mit IRQ 4
;              11 ... 15  dto.                          mit IRQ 3
;              21 ... 25  dto.                          mit IRQ 5
;              31 ... 35  dto.                          mit IRQ 9
;
;    n = 1: 3F8H,2F8H,3E8H,2E8H,280H,288H,290H,298H als COM1 - COM8
;        2: 2F8H,2A8H,2B0H,2B8H,1A9H,1A8H,1B0H,1B8H als COM2 - COM9
;        3: 2A0H,2A8H,2B0H,2B8H,1A0H,1A8H,1B0H,1B8H als COM3 - COM10
;        4: 160H,168H,170H,178H,180H,188H,190H,198H als COM3 - COM10
;        5: 3F8H,2F8H,2B0H,2B8H,1A0H,1A8H,1B0H,1B8H als COM1 - COM8
;
;  Schnittstellen-Protokoll:
;    Senden: Die Funktion 0 oder 4 (Modem an) gibt die
;       Signale DTR (pin 20) und RTS (pin 4) aus (+ 12 V).
;       Vor dem Senden eines Zeichens werden DSR (pin 6) und CTS (pin 5)
;       abgefragt. Sie mssen beide gesetzt sein (+12 V). Wenn diese
;       Bedingung nicht erfllt ist, erfolgt ein Rcksprung aus der
;       Senderoutine (Funktion 1) mit dem Time-out-bit (ah-bit 7)
;       auf 1 gesetzt.
;    Empfangen: Es ist grundstzlich DTR (pin 20) gesetzt. Alle ankommenden
;       Zeichen werden im Buffer zwischengespeichert (sofern mit Funktion 0
;       oder 4 das Modem eingeschaltet wurde).
;       Der Buffer kann max. 256 Zeichen pro Schnittstelle mit dem
;       zugehrigen Fehlerstatus (Overrun-, Parity- oder Framing-Error)
;       aufnehmen. Weitere Zeichen werden ignoriert.
;
;  Status-Abfrage:
;       Die Funktion 3 liest den gleichen Status wie im Original-
;       BIOS mit Ausnahme des DR-Bits (Data Ready, ah-bit 0). Dieses Bit
;       zeigt hier an, da im Input-Buffer mindestens ein Zeichen gespeichert
;       ist, das mit der Funktion 2 abgeholt werden kann.
;
;*  J. Rathlev, Apr. 1990
;*
;* 2.3          New Interrupt Routine: J. Rathlev, Sept. 1991
;*
;* 2.4          Supports COM1-8 or COM2-9: J. Rathlev, Dec. 1991
;
;* 3.1          Interrupt buffering for input/output: J. Rathlev, Feb. 1995
;  3.2          with multiple loop in interrupt routine
;  3.3          line counter for input buffer
;-------------------------------------------------------------------------

code    segment
        assume  cs:code
        assume  ds:code
        assume  es:nothing
        assume  ss:nothing

; BIOS Interrupt Number

BioInt  equ     14h

; Number of COM-channels

ComAnz  equ     8

CR      equ     0DH     ;carriage return

; Registers of 8250:

IER     equ     1       ; Interrupt Enable Register
IIR     equ     2       ; Interrupt Identification Register
LCR     equ     3       ; Line Control Register
MCR     equ     4       ; MODEM Control Register
LSR     equ     5       ; Line Status Register
MSR     equ     6       ; MODEM Status Register

EOI     equ     20h     ; end of interrupt

;-------------------------------------------------------------------------
        org     100h

start:  jmp     init

DrvName : db 'PC32DRVN.COM (3.31)',0
CopRgt  : db 'PC-COM(8)-Driver (Vers. 3.31 - (c) J.Rathlev 1995) $'

; input buffer

RcvLen    equ     256           ;receiver buffer length
SndLen    equ     512           ;send buffer length

RcvBuf  dw      ComAnz*RcvLen dup (?)
SndBuf  db      ComAnz*SndLen dup (?)

; bufferpointer and counter
; receive buffer
RcvBPtr  label   word
        n = 0
rept    ComAnz
        dw      RcvBuf+n
        n = n + 2*RcvLen
        endm

RcvCnt    dw      ComAnz dup (?)       ;buffer count
LineCnt   dw      ComAnz dup (?)       ;line count
RcvIPnt   dw      ComAnz dup (?)       ;input buffer pointer
RcvOPnt   dw      ComAnz dup (?)       ;output buffer pointer

; send buffer
SndBPtr  label   word
        n = 0
rept    ComAnz
        dw      SndBuf+n
        n = n + SndLen
        endm

SndCnt    dw      ComAnz dup (?)       ;buffer count
SndIPnt   dw      ComAnz dup (?)       ;input buffer pointer
SndOPnt   dw      ComAnz dup (?)       ;output buffer pointer

PortAdr   dw      ComAnz dup (?)       ;RS232 port Adr.

FrstCom   dw      ?                    ;first supp. Com-Nr.
LastCom   dw      ?                    ;first not supp. Com-Nr.

; interrupt masks and pointer

HwMsk     db      ?       ;Interrupt Mask
HwInt     db      ?       ;IRQ Number
HwiPtr    dd      ?       ;Pointer to Hardware Int.
IntPtr    dd      ?       ;Pointer to old BIOS-INT14
ComDrvID  dw      0F833h  ;driver identifier byte


;-----------------------------------------------------------------
; the following routine "ComRout" is the new INT14 for COM1
; port number in dx (0 ... ComAnz-1)
; function in ah:
;-------------------------------------------------------------------------
;  0    initialize the comm. port
;       enable interrupts from line 4
;       al contains parameter for init.:
;       normal: (bit 1 = 1)
;       bit
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -    110         X0 - none    0 - 1   10 - 7 bits
;       001 -    150         01 - odd     1 - 2   11 - 8 bits
;       010 -    300         11 - even
;       011 -    600
;       100 -   1200
;       101 -   2400
;       110 -   4800
;       111 -   9600
;       on return status in ah,al as function 3
;
;       extended: (bit 1 = 0)
;       7      6      5      4      3      2      1      0
;       -- baud rate --      -Parity-    stopbit  word length
;       000 -     50         X0 - none    0 - 1   00 - 7 bits
;       001 -     75         01 - odd     1 - 2   01 - 8 bits
;       010 -    134.5       11 - even
;       011 -   1800
;       100 -  19200
;       101 -  38400
;       110 -  57600
;       111 - 115200
;
;       on return status in ah,al as function 3
;
;-------------------------------------------------------------------------
;  1    send a char. in al to comm. line
;       al is saved
;       return with
;          ah bit 7 = 1 : unable to send char. to line
;                   = 0 : bit 0-6 contain line status
;
;-------------------------------------------------------------------------
;  2    read a char. to al from comm. line via buffer
;       return with
;          ah bit 7 = 1 : no char. in buffer
;                   = 0 : bit 1,2,3,4 contain line status
;
;-------------------------------------------------------------------------
;  3    read status from comm. port
;       return with
;          ah bit 7 : always 0
;                 6 : transmitter shift reg. empty
;                 5 : transmitter holding reg. empty
;                 4 : break interrupt
;                 3 : framing error
;                 2 : parity error
;                 1 : overrun error
;                 0 : data received (in buffer)
;          al bit 7 : received line signal detect
;                 6 : ring indicator
;                 5 : data set ready
;                 4 : clear to send
;                 3 : delta received line signal detect
;                 2 : trailing edge ring indicator
;                 1 : delta data set ready
;                 0 : delta clear to send
;
;-------------------------------------------------------------------------
;  4    switch modem on (set DTR and RTS, enable interrupts)
;
;-------------------------------------------------------------------------
;  5    switch modem off (reset DTR and RTS, interrupt disable)
;
;-------------------------------------------------------------------------
;  6    read line control register to al
;-------------------------------------------------------------------------
;  7    read buffer counter to ax, buffer length to dx
;-------------------------------------------------------------------------
;  8    driver identifikation: set al=0 on call
;                              PC32-driver installed:
;                                     return al=$F8
;                              otherwise al unchanged
;-------------------------------------------------------------------------

ComRout:sti
        push    ds
        push    si
        push    di
        push    cs
        pop     ds              ;copy cs to ds
        cmp     dx,FrstCom      ;number of 1.COM
        jl      Direct          ;not supp. COM-number
        cmp     dx,LastCom      ;number of first not supp. COM-number
        jge     Direct          ;not supp. COM-number
        push    bx
        push    cx
        push    dx
        sub     dx,FrstCom      ;subtract Offset
        mov     bl,ah           ;function to bl
        xor     bh,bh
        shl     bx,1            ;*2
        cmp     bx,FctNr        ;legal function?
        jnc     ComFin          ;no, return imm.
        shl     dx,1
        mov     si,dx           ;2* rel. ComNr.
        mov     dx,[si+PortAdr] ;Portaddr.
        jmp     word ptr [bx+offset FctTab]   ;branch

FctTab  dw      ComSet          ;set baud rate and format
        dw      ComXmit         ;send char. via ser. port
        dw      ComRecv         ;get a received char. from buffer
        dw      ComStat         ;get modem status
        dw      ComOn           ;switch modem on
        dw      ComOff          ;switch modem off
        dw      ComLCR          ;read LCR
        dw      ComBCnt         ;read buffer counter
        dw      ComId           ;identify driver
FctNr   equ     $-FctTab

;-------------------------------------------------------------------------
ComFin: pop     dx
        pop     cx
        pop     bx
ComRet: pop     di
        pop     si
        pop     ds
        iret

Direct: pushf                   ;prepare to emulate Int-Call
        cli
        call    dword ptr [IntPtr]
        jmp     ComRet

;-------------------------------------------------------------------------
;baud rate table

BTab1   dw      1047            ;110 baud
        dw      768             ;150 baud
        dw      384             ;300 baud
        dw      192             ;600 baud
        dw      96              ;1200 baud
        dw      48              ;2400 baud
        dw      24              ;4800 baud
        dw      12              ;9600 baud
BTab2   dw      2304            ;50 baud
        dw      1536            ;75 baud
        dw      857             ;134.5 baud
        dw      64              ;1800 baud
        dw      6               ;19200 baud
        dw      3               ;38400 baud
        dw      2               ;57600 baud
        dw      1               ;115200 baud

;-------------------------------------------------------------------------
;set ser. port to new baud rate and format in ax
;si = 2*port nr., dx = port addr., ds = cs
;new values in al

ComSet: push    ax              ;save
        call    ComRlse
        pop     ax
        push    dx              ;save port addr.
        mov     ah,al           ;save
        add     dx,LCR          ;line control reg.
        mov     al,80H          ;set DLAB
        out     dx,al
        mov     bl,ah
        mov     cl,4            ;rot count
        rol     bl,cl           ;rotate 4 bits
        and     bx,0EH          ;mask out baud rate
        test    ah,2            ;extended mode?
        jz      ComExt
        add     bx,offset BTab1 ;pointer to 1. divisor table
        jmp     short ComNor
ComExt: add     bx,offset BTab2 ;pointer to 2. divisor table
ComNor: dec     dx
        dec     dx              ;divisor latch (DLM)
        mov     al,[bx+1]
        out     dx,al
        dec     dx              ;divisor latch (DLL)
        mov     al,[bx]
        out     dx,al
        add     dx,LCR          ;line control reg.
        mov     al,ah           ;format bits
        or      al,2            ;set bit 1
        and     al,1FH          ;mask out
        out     dx,al
        pop     dx              ;restore port base addr.
        call    ComInit         ;initialize
        jmp     ComStat


;-------------------------------------------------------------------------
;wait until char. received and read char. from buffer
;si = 2*port nr., dx = port addr., ds = cs
;return with char. in al, ah = 80h on time out, else = Line Status

ComRecv:lea     bx,[si+RcvCnt]    ;test buffer count
        xor     cx,cx
ComRWt: cmp     word ptr [bx],0
        jnz     ComRc2          ;char. received
        loop    ComRWt          ;wait until not zero
        mov     ah,80H        ;time out
        jmp     ComFin          ;return imm.

ComRc2: dec     word ptr [bx]   ;count one down
        mov     bx,[si+RcvOPnt]   ;buffer pointer
        mov     di,[si+RcvBPtr]   ;buffer base addr.
        mov     ax,[bx+di]      ;get status and char. from buffer
        inc     bx              ;incr. buffer pointer
        inc     bx
        cmp     bx,2*RcvLen       ;buffer length
        jc      ComRc1
        xor     bx,bx           ;set to first buffer addr.
ComRc1: mov     [si+RcvOPnt],bx   ;and store
        cmp     al,CR           ;carriage return?
        jnz     ComRc3
        cmp     word ptr [si+LineCnt],0
        jz      ComRc3
        dec     word ptr [si+LineCnt]
ComRc3: jmp     ComFin


;-------------------------------------------------------------------------
;send char. in al via ser. port
;si = 2*port nr., dx = port addr., ds = cs
;return with ah (bit 7) = 1 buffer full, al = char
;            ah (bit 7) = 0 on success, al =char

ComXmit:lea     bx,[si+SndCnt]    ;address of buffer counter
        cmp     word ptr [bx],SndLen
        jge     ComXm1            ;buffer full

        cli
        or      word ptr [bx],0
        jnz     ComXm3            ;interrupt active
        mov     ah,al             ;save char.
        add     dx,IER
        call    SndEna            ;enable THRE interr.
        mov     al,ah
ComXm3: inc     word ptr [bx]     ;incr. buffer count
        mov     bx,[si+SndIPnt]   ;buffer pointer
        mov     di,[si+SndBPtr]   ;buffer base addr.
        mov     [bx+di],al        ;store to buffer
        inc     bx                ;incr. buffer pointer
        cmp     bx,SndLen         ;buffer length
        jc      ComXm2
        xor     bx,bx             ;set to first buffer addr.
ComXm2: mov     [si+SndIPnt],bx   ;and store
        sti
        xor     ah,ah
        jmp     comfin

ComXm1: mov     ah,80H          ;set error bit
        jmp     ComFin


;-------------------------------------------------------------------------
;read status from ser. port
;si = 2*port nr., dx = port addr., ds = cs
;ah and al contain the status bits like the orig. BIOS-Int.
;except bit 0 in ah (data ready). This bit is set if the input buffer
;is not empty (RcvCnt)

ComStat:add     dx,LSR          ;line status reg.
        in      al,dx
        mov     ah,al
        inc     dx              ;modem status reg.
        in      al,dx
        and     ah,0FEH         ;mask out bit 0
        mov     bx,[si+RcvCnt]
        or      bx,bx
        jz      ComSt1
        or      ah,1            ;set bit 0
ComSt1: jmp     ComFin


;-------------------------------------------------------------------------
;switch modem on
;si = 2*port nr., dx = port addr., ds = cs

ComOn:  call    ComInit
        jmp     ComFin


;-------------------------------------------------------------------------
;switch modem off
;si = 2*port nr., dx = port addr., ds = cs

ComOff: call    ComRlse
        jmp     ComFin


;-------------------------------------------------------------------------
; read line control register
;si = 2*port nr., dx = port addr., ds = cs

ComLCR: add     dx,LCR
        in      al,dx
        jmp     ComFin


;-------------------------------------------------------------------------
; read buffer counter
;si = 2*port nr., dx = port addr., ds = cs
; al = 0 (Rcv)
;      1 (Snd)
; return: ax = buffer counter
;         cx = line counter (on receive mode)
;         dx = buffer length

ComBCnt:pop     dx
        pop     cx
        pop     bx
        or      al,al
        jz      GetRcv
        mov     ax,[si+SndCnt]
        mov     dx,SndLen
        jmp     ComRet

GetRcv: mov     ax,[si+RcvCnt]
        mov     cx,[si+LineCnt]
        mov     dx,RcvLen
        jmp     ComRet


;-------------------------------------------------------------------------
; get driver identifier byte
; al = $F8 for PC32-driver
; al = version
; cl = IRQ
; dx = port
; es:bx = pointer to driver name

ComID:  mov     ax,dx           ;save port addr.
        pop     dx
        pop     cx
        pop     bx
        lea     bx,[DrvName]
        mov     dx,ax
        mov     ax,cs
        mov     es,ax
        mov     cl,[HwInt]      ;interrupt
        mov     ax,ComDrvID     ;identification byte
        jmp     ComRet


;-------------------------------------------------------------------------
;subroutine test ser. port status
;in dx port addr.of status reg., in bh status bits to test
;destroyed: cx
;return with zero flag = 0 if time out, status in ah

TstStat:xor     cx,cx           ;clear
TstSta1:in      al,dx
        mov     ah,al
        xor     al,0FFH         ;complement
        test    al,bl           ;test bits
        jz      TstStaR         ;o.k. return
        loop    TstSta1         ;try again
        or      bh,bh           ;clear zero flag
TstStaR:ret


;-------------------------------------------------------------------------
;subroutine init. buffer and set modem on
;si = 2*port nr., dx = base port addr.; ds = cs

ComInit:push    dx              ;save port-addr.
        xor     ax,ax
        cli
        mov     [si+RcvCnt],ax    ;buffer count
        mov     [si+LineCnt],ax   ;         line counter
        mov     [si+RcvIPnt],ax   ;buffer pointer
        mov     [si+RcvOPnt],ax
        mov     [si+SndCnt],ax    ;send buffer count
        mov     [si+SndIPnt],ax   ;     buffer pointer
        mov     [si+SndOPnt],ax
        in      al,dx           ;clear receiver buffer reg.
        add     dx,MCR          ;modem control reg. - modem on
        mov     al,0BH          ;set DTR, RTS and OUT2
        out     dx,al
        sub     dx,MCR-IER      ;enable interrupt
        mov     al,1            ;for received and send
        out     dx,al
        mov     ah,HwMsk        ;set interrupt controller
        not     ah
        mov     al,HwInt
        mov     dx,21h          ;1. int. contr.
        test    al,8            ;2. ?
        jz      SetMsk
        mov     dx,0A1h         ;2. int. contr.
SetMsk: in      al,dx           ;interrupt mask
        and     al,ah           ;enable interrupt
        out     dx,al
        sti
        pop     dx
        ret


;-------------------------------------------------------------------------
;subroutine release and modem off
;dx = base port addr.; ds = cs

ComRlse:
        push    dx              ;save port-addr.
        xor     al,al
        out     dx,al           ;IER
        add     dx,3            ;MCR - set modem off
        out     dx,al
        mov     ah,HwMsk
        mov     al,HwInt
        mov     dx,21h          ;1. int. contr.
        test    al,8            ;2. ?
        jz      ClrMsk
        mov     dx,0A1h         ;2. int. contr.
ClrMsk: in      al,dx           ;interrupt mask
        or      al,ah           ;disable interrupt
        out     dx,al
        pop     dx
        ret


;-------------------------------------------------------------------------
;enable interrupts from THRE
; dx = port addr. of IER

SndEna: in      al,dx
        or      al,2              ;enable THRE-Interr.
        out     dx,al
        ret

;-------------------------------------------------------------------------
;disable interrupt from THRE
; dx = port addr. of IER

SndDis: in      al,dx
        and     al,0Dh          ;disable THRE-Interr.
        out     dx,al
        ret

;-------------------------------------------------------------------------
;receive char. interrupt
; si =  2 * rel. ComNr
; dx =  portaddr. + IIR

assume ds:code

RcvInt: add     dx,LSR-IIR      ;Line status reg.
        in      al,dx
        and     al,1EH          ;mask out error bits
        mov     ah,al
        sub     dx,LSR
        in      al,dx           ;read char. from ser.port

        lea     bx,[si+RcvCnt]            ;address of buffer counter
        cmp     word ptr [bx],RcvLen
        jz      RcvFul                    ;buffer full

        cmp     al,CR                     ;carriage return
        jnz     RcvI1
        inc     word ptr [si+LineCnt]     ;incr. line count
RcvI1:  inc     word ptr [bx]             ;incr. buffer count
        mov     bx,[si+RcvIPnt]           ;buffer pointer
        mov     di,[si+RcvBPtr]           ;buffer base addr.
        mov     [bx+di],ax                ;store to buffer
        inc     bx                        ;incr. pointer
        inc     bx
        cmp     bx,2*RcvLen
        jc      StBiP
        xor     bx,bx                     ;init. buffer addr.
StBiP:  mov     [si+RcvIPnt],bx           ;store

RcvFul: add     dx,IIR
        ret

;send char. interrupt

SndInt: or      word ptr [si+SndCnt],0  ;buffer counter
        jz      SndOff                  ;buffer not busy
        dec     word ptr [si+SndCnt]    ;decr. buffer count
        sub     dx,IIR
        mov     bx,[si+SndOPnt]         ;buffer pointer
        mov     di,[si+SndBPtr]         ;buffer base addr.
        mov     al,[bx+di]              ;load from buffer
        out     dx,al                   ;send char.
        inc     bx                      ;incr. pointer
        cmp     bx,SndLen
        jc      StBoP
        xor     bx,bx                   ;init. buffer addr.
StBoP:  mov     [si+SndOPnt],bx         ;store
        add     dx,IIR
        ret

SndOff: dec     dx              ;to IER
SoOk:   call    SndDis
        inc     dx              ;to IIR
        ret


; MSR interrupt, should not happen

MSRInt: add     dx,MSR-IIR
        in      al,dx
        sub     dx,MSR-IIR
        ret

; LSR interrupt, should not happen

LSRInt: add     dx,LSR-IIR
        in      al,dx
        sub     dx,LSR-IIR
        ret

assume ds:nothing

;-------------------------------------------------------------------------
;interrupt routine for ser. port (int 0CH/0BH/0DH)

;entry point

ComInt: sti
        push    ax
        push    bx
        push    cx
        push    dx
        push    di
        push    si
        push    ds

        mov     ax,cs
        mov     ds,ax
        mov     cx,5            ;try 5 times

assume ds:code

IntLp:  xor     si,si           ;= 0
        push    cx
        mov     cx,ComAnz       ;max. number of ser. ports

;Test next COM

TstInt: mov     dx,[si+PortAdr]         ;get port address from table
        add     dx,IIR
IPoll:  in      al,dx           ;read IIR
        test    al,1            ;interrupt pending?
        jnz     NxtInt          ;no
        and     ax,06h
        mov     bx,ax
        call    word ptr [bx+offset IntBra]
        jmp     short IPoll

NxtInt: inc     si
        inc     si                      ;si = 2* rel. ComNr
        loop    TstInt                  ;try next
        pop     cx
        loop    IntLp           ;try again

        pop     ds              ;restore registers
        pop     si
        pop     di
        pop     dx
        pop     cx
        pop     bx
        mov     al,HwInt
        cli
        test    al,8
        mov     al,EOI          ;end of  interrupt
        jz      EOInt1          ;1. contr.
        out     0A0h,al         ;end of  interrupt (contr.2)
        jcxz    EOInt1          ;wait state
EOInt1: out     20h,al          ;end of  interrupt (contr.1)
        pop     ax
        iret

IntBra: dw      MSRInt
        dw      SndInt
        dw      RcvInt
        dw      LSRInt

DrvEnd:

;============================================================================
;Laden oder Entfernen des Treibers fr die ser. Schnittstellen
;Startpunkt des .COM-Programms
;Der folgende Programmteil bleibt nach dem Ende des Programms
;nicht im Speicher
;Command-String:  PC32DrvN    ==>  COM1-Treiber installieren
;                 PC32DrvN /n ==>  COMn-Treiber installieren (n=1,..)
;                 PC32DrvN /u ==>  COM-treiber wieder entfernen

        ;Interrupt Masken

Maske   db      10h      ;Int. 4
        db      8        ;Int. 3
        db      20h      ;Int. 5
        db      2        ;Int. 9

IntAnz = Intad - Maske   ;Anzahl der mglichen Interrupts

        ;Interrupt Adressen

IntAd   db      4
        db      3
        db      5
        db      9

        ;Bildschirmhinweise

Mess1   db      'with IRQ'
IRQNr   db      '4 installed!',13,10,'$'
Mess2   db      'removed!',13,10,'$'
Mess3   db      'already installed!',13,10,'$'
Mess4   db      'not installed!',13,10,'$'
Mess5   db      13,10,'Calling: PC23DRVN /Mode',13,10
        db      ' Mode:   COM1..8  COM2..9  COM3..10 COM3..10 COM1..8',13,10
        db      '  IRQ 4:    1 (def)  2        3        4        5',13,10
        db      '  IRQ 3:   11       12       13       14       15',13,10
        db      '  IRQ 5:   21       22       23       24       25',13,10
        db      '  IRQ 9:   31       32       33       34       35',13,10,'$'


        ;Programmanfang

Init:   mov     ax,cs              ;copy cs to ss
        mov     ss,ax
        mov     sp,offset Tops     ;init stack pointer

        mov     dx,offset CopRgt
        call    OMess              ;Hinweis auf Copyright ausgeben

        mov     bx,80h             ;Command-String lesen
        mov     ch,0
        mov     cl,[bx]            ;Lnge
        or      cl,cl
        jz      Standrd            ;kein Comm-String

RdPar:  inc     bx
        mov     al,[bx]
        cmp     al,'/'             ;Zeichen prfen
        jz      TstOpt             ;Option prfen
        cmp     al,'?'
        jz      Error
        loop    RdPar

Standrd:mov     dx,1               ;setze auf Mode 1
        jmp     Install

;-------------------------------------------------------------------------
Error:  mov     dx,offset Mess5
        jmp     Ende

;-------------------------------------------------------------------------
UnInst: call    GetInt
        jnz     Unin1
        mov     ax,es:[bx+ComDrvID-ComRout]
        cmp     ax,ComDrvID     ;Treiberidentifizierung
        jnz     UnIn1
        jmp     UnIn0           ;Installation entfernen

UnIn1:  mov     dx,offset Mess4 ;war nicht installiert
        jmp     Ende

;-------------------------------------------------------------------------
TstOpt: dec     cx
        mov     al,[bx+1]       ;nchstes Zeichen prfen
        cmp     al,'u'
        jz      UnInst          ;Treiber entfernen
        cmp     al,'U'
        jz      UnInst          ;dto.
        call    TstDig          ;Ziffer?
        jz      RdPar           ;nein, nchstes Zeichen
        xor     ah,ah           ;= 0
        mov     dx,ax
        dec     cx
DigLp:  inc     bx
        mov     al,[bx+1]       ;nchstes Zeichen
        call    TstDig
        jz      Install         ;keine Ziffer mehr
        call    MulAd10
        loop    DigLp

Install:mov     cx,IntAnz       ;Anzahl der mglichen Interrupts
        xor     al,al           ;= 0
ITest:  cmp     dx,10           ;unter 10?
        jbe     Inst0
        inc     al
        sub     dx,10           ;nchster Interrupt
        loop    ITest
        jmp     Error           ;falsche Angabe

Inst0:  cmp     dx,0
        jbe     Error           ;falscher Mode
        cmp     dx,ModeAnz      ;max. Anzahl von Modes
        ja      Error           ;falscher Mode
        shl     dx,1            ;*2
        mov     si,dx
        mov     si,cs:[si+ModeTab-2] ;Tabelle mit Portadressen
        mov     di,offset PortAdr    ;Zieltabelle
        mov     cx,ComAnz
        cld
        rep     movsw                ;Tabelle nach PortAdr kopieren
        mov     cx,[si]              ;first COM-Nr. supp.
        mov     FrstCom,cx
        add     cx,ComAnz
        mov     LastCom,cx

        xor     ah,ah
        mov     bx,ax
        mov     ah,[offset Maske + bx]  ;Interrupt Maske
        mov     HwMsk,ah
        mov     ah,[offset IntAd + bx]  ;Interrupt Adresse
        mov     HwInt,ah
        add     ah,30h
        mov     IRQNr,ah           ;IRQ-Nr. fr Meldung

        call    GetInt             ;alte Interrupt-Adr. prfen
        jnz     Inst1              ;sind ungleich ==> installieren
        mov     dx,offset Mess3
        jmp     Ende               ;sind gleich ==> nichts tun

;-------------------------------------------------------------------------
Inst1:  cli
        mov     word ptr IntPtr,bx   ;alten Interrupt merken
        mov     word ptr IntPtr+2,es
        mov     ax,2500h+BioInt      ;neuen Interrupt setzen
        int     21h

        mov     ah,35h
        mov     al,HwInt
        test    al,8                 ;2. Controller?
        jz      SvH1
        add     al,68h
        jmp     short SvH2
SvH1:   add     al,8
SvH2:   int     21h                       ;alten Hardware Int.
        mov     word ptr HwiPtr,bx   ;merken
        mov     word ptr HwiPtr+2,es

        mov     dx,offset ComInt     ;neuen Hardware Interr.
        mov     ah,25h               ;setzen
        int     21h

        mov     cx,ComAnz
        mov     dx,FrstCom           ;Com-Nr.
        call    AllOff
        sti
        mov     dx,offset Mess1
        call    OMess            ;Hinweis ausgeben

        mov     dx,offset DrvEnd ;Das Program ist bis "DrvEnd"
                                 ;speicherresident
        int     27h              ;Ende


;-------------------------------------------------------------------------
;alle Coms ausschalten
;cx = Anzahl Coms, dx = 1. ComNr

AllOff: mov     ah,5            ;alle COMs aus
        int     14h
        inc     dx
        loop    AllOff
        ret

;-------------------------------------------------------------------------
UnIn0:  mov     cx,es:[bx+LastCom-ComRout]
        mov     dx,es:[bx+FrstCom-ComRout]         ;Com-Nr.
        sub     cx,dx
        inc     cx                 ;number of COMs
        call    AllOff

        cli
        mov     dx,es:[bx+IntPtr-ComRout]
        mov     ds,es:[bx+IntPtr+2-ComRout]
        mov     ax,2500h+BioInt    ;Original-BIOS-Adresse bei ComRout-4
        int     21h

        mov     dx,es:[bx+HwiPtr-ComRout]
        mov     ds,es:[bx+HwiPtr+2-ComRout]
        mov     ah,25h             ;Original-Adresse des Hardw.Int.
        mov     al,es:[bx+HwInt-ComRout]
        add     al,8
        int     21h

        sti
        mov     dx,offset Mess2
        mov     bx,cs
        mov     ds,bx
Ende:   call    OMess           ;Hinweis ausgeben
        mov     ah,0
        int     21h


; Bildschirmmeldung
; Pointer auf text in dx

OMess:  mov     ah,9            ;output a string to screen
        int     21h
        ret

;-------------------------------------------------------------------------
; alte INT14-Adresse holen

GetInt: mov     ax,3500h+BioInt    ;alte Int.-Adr. lesen
        int     21h
        mov     dx,offset ComRout  ;neue INT14-Adresse
        cmp     dx,bx              ;bereits installiert?
        ret

;-------------------------------------------------------------------------
; Zeichen in al auf Ziffer prfen
; Zero Flag = 1: nein
;           = 0: ja

TstDig: cmp     al,30h          ;Ziffer
        jb      TstDigN         ;no
        cmp     al,39h
        ja      TstDigN         ;no
        sub     al,30H          ;convert to binary and reset zero flag
        ret

TstDigN:xor     al,al           ;set zero flag
        ret

;-------------------------------------------------------------------------
; Multipliziere dx mit 10 und addiere ax

MulAd10:shl     dx,1            ;*2
        push    dx
        shl     dx,1            ;*4
        shl     dx,1            ;*8
        add     dx,ax
        pop     ax              ;dx*2
        add     dx,ax
        ret

;-------------------------------------------------------------------------
; Portadressen:

ModeTab dw      Mode1,Mode2,Mode3,Mode4,Mode5
ModeAnz equ     ($-ModeTab)/2

Mode1   dw      3F8H,2F8H,3E8H,2E8H,280H,288H,290H,298H,0
Mode2   dw      2F8H,2A8H,2B0H,2B8H,1A9H,1A8H,1B0H,1B8H,1
Mode3   dw      2A0H,2A8H,2B0H,2B8H,1A0H,1A8H,1B0H,1B8H,2
Mode4   dw      160H,168H,170H,178H,180H,188H,190H,198H,2
Mode5   dw      3F8H,2F8H,2B0H,2B8H,1A0H,1A8H,1B0H,1B8H,0

;-------------------------------------------------------------------------
;space for user stack
        dw      100h dup (?)
Tops    label word

code    ends
        end     start

